# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from typing import Dict, List

from rp2.configuration import Configuration
from rp2.gain_loss import GainLoss
from rp2.gain_loss_set import GainLossSet
from rp2.in_transaction import InTransaction
from rp2.intra_transaction import IntraTransaction
from rp2.out_transaction import OutTransaction
from rp2.plugin.accounting_method.fifo import AccountingMethod
from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError
from rp2_test_output import RP2_TEST_OUTPUT  # pylint: disable=wrong-import-order

_ASSETS: List[str] = ["B1", "B2", "B3", "B4"]


class TestGainLossSet(unittest.TestCase):
    _configuration: Configuration
    _accounting_method: AccountingMethod

    _in3: Dict[str, InTransaction]
    _in2: Dict[str, InTransaction]
    _in6: Dict[str, InTransaction]
    _in5: Dict[str, InTransaction]
    _in4: Dict[str, InTransaction]

    _out15: Dict[str, OutTransaction]
    _out14: Dict[str, OutTransaction]
    _out16: Dict[str, OutTransaction]
    _out12: Dict[str, OutTransaction]
    _out13: Dict[str, OutTransaction]

    _intra25: Dict[str, IntraTransaction]
    _intra24: Dict[str, IntraTransaction]
    _intra22: Dict[str, IntraTransaction]

    _gain_loss1: GainLoss
    _gain_loss2: GainLoss

    _gain_loss_set: Dict[str, GainLossSet]

    @classmethod
    def setUpClass(cls) -> None:
        cls._configuration = Configuration("./config/test_data.config", US())
        cls._accounting_method = AccountingMethod()

        cls._in3 = {}
        cls._in2 = {}
        cls._in6 = {}
        cls._in5 = {}
        cls._in4 = {}

        cls._out15 = {}
        cls._out14 = {}
        cls._out16 = {}
        cls._out12 = {}
        cls._out13 = {}

        cls._intra25 = {}
        cls._intra24 = {}
        cls._intra22 = {}

        cls._gain_loss_set = {}

        for asset in _ASSETS:
            cls._initialize_transactions(asset)

    @classmethod
    # Reproduce programmatically the data from sheets B1 to B4 in input/test_data.ods
    def _initialize_transactions(cls, asset: str) -> None:
        cls._in3[asset] = InTransaction(
            cls._configuration,
            "2020-01-01 08:41:00.000000 +0000",
            asset,
            "Coinbase",
            "Bob",
            "BUY",
            RP2Decimal("11000"),
            RP2Decimal("1"),
            RP2Decimal("100.00"),
            unique_id=4,
        )
        cls._in2[asset] = InTransaction(
            cls._configuration,
            "2020-02-01 11:18:00.000000 +0000",
            asset,
            "BlockFi",
            "Bob",
            "INTEREST",
            RP2Decimal("12000.0"),
            RP2Decimal("2.0"),
            RP2Decimal("0"),
            unique_id=3,
        )
        cls._in6[asset] = InTransaction(
            cls._configuration,
            "2020-03-01 09:45:00.000000 +0000",
            asset,
            "BlockFi",
            "Bob",
            "INTEREST",
            RP2Decimal("13000.0"),
            RP2Decimal("3"),
            RP2Decimal("0"),
            unique_id=7,
        )
        cls._in5[asset] = InTransaction(
            cls._configuration,
            "2020-04-01T09:45Z",
            asset,
            "Coinbase",
            "Bob",
            "BUY",
            RP2Decimal("14000.0"),
            RP2Decimal("4.0"),
            RP2Decimal("400"),
            unique_id=6,
        )
        cls._in4[asset] = InTransaction(
            cls._configuration,
            "2020-05-01T14:03Z",
            asset,
            "Coinbase",
            "Bob",
            "BUY",
            RP2Decimal("15000.0"),
            RP2Decimal("5.0"),
            RP2Decimal("500"),
            unique_id=5,
        )
        cls._out15[asset] = OutTransaction(
            cls._configuration,
            "2020-01-11 11:15:00.000000 +0000",
            asset,
            "Coinbase",
            "Bob",
            "SELL",
            RP2Decimal("11200.0"),
            RP2Decimal("0.2"),
            RP2Decimal("0"),
            unique_id=16,
        )
        cls._out14[asset] = OutTransaction(
            cls._configuration,
            "2020-02-11 19:58:00.000000 +0000",
            asset,
            "Coinbase",
            "Bob",
            "SELL",
            RP2Decimal("12200.0"),
            RP2Decimal("1.0"),
            RP2Decimal("0"),
            unique_id=15,
        )
        cls._out16[asset] = OutTransaction(
            cls._configuration,
            "2020-04-11 07:10:00.000000 +0000",
            asset,
            "BlockFi",
            "Bob",
            "GIFT",
            RP2Decimal("14200.00"),
            RP2Decimal("5.0"),
            RP2Decimal("0"),
            unique_id=17,
        )
        cls._out12[asset] = OutTransaction(
            cls._configuration,
            "2020-04-12 17:50:00.000000 +0000",
            asset,
            "Coinbase",
            "Bob",
            "DONATE",
            RP2Decimal("14300"),
            RP2Decimal("3.79"),
            RP2Decimal("0"),
            unique_id=13,
        )
        cls._out13[asset] = OutTransaction(
            cls._configuration,
            "2021-06-11 05:31:00.000000 +0000",
            asset,
            "Coinbase",
            "Bob",
            "SELL",
            RP2Decimal("20200.00"),
            RP2Decimal("2"),
            RP2Decimal("0.01"),
            unique_id=14,
        )
        cls._intra25[asset] = IntraTransaction(
            cls._configuration,
            "2020-01-21 18:33:14.342000 +0000",
            asset,
            "Coinbase",
            "Bob",
            "BlockFi",
            "Bob",
            RP2Decimal("11400.0"),
            RP2Decimal("0.1"),
            RP2Decimal("0.09"),
            unique_id=26,
        )
        cls._intra24[asset] = IntraTransaction(
            cls._configuration,
            "2020-05-21 12:58:10.000000 +0000",
            asset,
            "Coinbase",
            "Bob",
            "Kraken",
            "Alice",
            RP2Decimal("14400.0"),
            RP2Decimal("0.2"),
            RP2Decimal("0.18"),
            unique_id=25,
        )
        cls._intra22[asset] = IntraTransaction(
            cls._configuration,
            "2021-07-21 10:02:02.000000 +0000",
            asset,
            "Coinbase",
            "Bob",
            "Kraken",
            "Alice",
            RP2Decimal("21400.0"),
            RP2Decimal("0.5"),
            RP2Decimal("0.46"),
            unique_id=23,
        )
        cls._gain_loss_set[asset] = GainLossSet(cls._configuration, cls._accounting_method, asset)
        if asset == "B1":
            # In transactions only
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("2"), cls._in2[asset], None))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("3"), cls._in6[asset], None))
        elif asset == "B2":
            # In and out transactions only
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.2"), cls._out15[asset], cls._in3[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("2"), cls._in2[asset], None))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.8"), cls._out14[asset], cls._in3[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.20"), cls._out14[asset], cls._in2[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("3"), cls._in6[asset], None))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("1.8"), cls._out16[asset], cls._in2[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("3"), cls._out16[asset], cls._in6[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.2"), cls._out16[asset], cls._in5[asset]))
            cls._gain_loss_set[asset].add_entry(
                GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("3.79"), cls._out12[asset], cls._in5[asset])
            )  # ==
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.01"), cls._out13[asset], cls._in5[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("2.0"), cls._out13[asset], cls._in4[asset]))
        elif asset == "B3":
            # In and intra transactions only
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.01"), cls._intra25[asset], cls._in3[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("2"), cls._in2[asset], None))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("3"), cls._in6[asset], None))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.02"), cls._intra24[asset], cls._in3[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.04"), cls._intra22[asset], cls._in3[asset]))
        else:  # asset == "B4", "BTC" or others
            cls._gain_loss1 = GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.2"), cls._out15[asset], cls._in3[asset])
            cls._gain_loss2 = GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.01"), cls._intra25[asset], cls._in3[asset])
            cls._gain_loss_set[asset].add_entry(cls._gain_loss1)
            cls._gain_loss_set[asset].add_entry(cls._gain_loss2)
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("2"), cls._in2[asset], None))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.79"), cls._out14[asset], cls._in3[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.21"), cls._out14[asset], cls._in2[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("3"), cls._in6[asset], None))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("1.79"), cls._out16[asset], cls._in2[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("3"), cls._out16[asset], cls._in6[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.21"), cls._out16[asset], cls._in5[asset]))
            cls._gain_loss_set[asset].add_entry(
                GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("3.79"), cls._out12[asset], cls._in5[asset])
            )  # ==
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.02"), cls._intra24[asset], cls._in4[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("2.01"), cls._out13[asset], cls._in4[asset]))
            cls._gain_loss_set[asset].add_entry(GainLoss(cls._configuration, cls._accounting_method, RP2Decimal("0.04"), cls._intra22[asset], cls._in4[asset]))

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    # Reproduce programmatically the data from sheets B1 to B4 in input/test_data.ods
    # and check them against golden output
    def test_good_gain_loss_sets(self) -> None:
        for asset in _ASSETS:
            self.assertEqual(str(self._gain_loss_set[asset]), RP2_TEST_OUTPUT[asset])

    def test_bad_gain_loss_set(self) -> None:
        gain_loss_set: GainLossSet
        asset: str = "B4"
        gain_loss1: GainLoss = self._gain_loss1
        gain_loss2: GainLoss = self._gain_loss2
        in3: InTransaction = self._in3[asset]
        in5: InTransaction = self._in5[asset]
        in6: InTransaction = self._in6[asset]
        out15: OutTransaction = self._out15[asset]
        out14: OutTransaction = self._out14[asset]
        out16: OutTransaction = self._out16[asset]

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            GainLossSet(None, self._accounting_method, asset)  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            GainLossSet("configuration", self._accounting_method, asset)  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'accounting_method' is not of type AbstractAccountingMethod: .*"):
            # Bad configuration
            GainLossSet(self._configuration, None, asset)  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            # Bad asset
            GainLossSet(self._configuration, self._accounting_method, None)  # type: ignore

        with self.assertRaisesRegex(RP2ValueError, "Parameter 'asset' value is not known: .*"):
            # Bad configuration
            GainLossSet(self._configuration, self._accounting_method, "foobar")

        gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry' is not of type GainLoss: .*"):
            # Bad transaction add
            gain_loss_set.add_entry(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry' is not of type GainLoss: .*"):
            # Bad transaction add
            gain_loss_set.add_entry(1111)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry' is not of type GainLoss: .*"):
            # Bad transaction add
            gain_loss_set.add_entry(in3)
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry' is not of type GainLoss: .*"):
            # Bad get_parent parameter
            gain_loss_set.get_parent(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry' is not of type GainLoss: .*"):
            # Bad get_parent parameter
            gain_loss_set.get_parent(1111)  # type: ignore

        with self.assertRaisesRegex(RP2ValueError, "Unknown entry:.*"):
            # Unknown get_parent parameter (set empty)
            gain_loss_set.get_parent(gain_loss1)

        gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)
        gain_loss_set.add_entry(gain_loss1)

        with self.assertRaisesRegex(RP2ValueError, "Unknown entry:.*"):
            # Unknown get_parent parameter (set non-empty)
            gain_loss_set.get_parent(gain_loss2)
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'transaction' is not of type AbstractTransaction: .*"):
            # Bad get_taxable_event_fractions parameter
            gain_loss_set.get_taxable_event_number_of_fractions(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'transaction' is not of type AbstractTransaction: .*"):
            # Bad get_taxable_event_fractions parameter
            gain_loss_set.get_taxable_event_number_of_fractions(self)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Unknown transaction:.*"):
            # Bad get_taxable_event_fractions parameter
            gain_loss_set.get_taxable_event_number_of_fractions(out15)
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'transaction' is not of type InTransaction: .*"):
            # Bad get_taxable_event_fractions parameter
            gain_loss_set.get_acquired_lot_number_of_fractions(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'transaction' is not of type InTransaction: .*"):
            # Bad get_taxable_event_fractions parameter
            gain_loss_set.get_acquired_lot_number_of_fractions(self)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Unknown transaction:.*"):
            # Bad get_taxable_event_fractions parameter
            gain_loss_set.get_acquired_lot_number_of_fractions(in3)
        with self.assertRaisesRegex(
            RP2ValueError, "Timestamp .* of acquired_lot entry .* is incompatible with timestamp .* of its ancestor .* using .* accounting method: .*"
        ):
            gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)
            in_transaction_test: InTransaction = InTransaction(
                self._configuration,
                "2020-01-05 11:18:00.000000 +0000",
                asset,
                "Coinbase",
                "Bob",
                "INTEREST",
                RP2Decimal("12000.0"),
                RP2Decimal("2.0"),
                RP2Decimal("0"),
                unique_id=3,
            )
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.2"), out15, in_transaction_test))
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.01"), out14, in3))
            for _ in gain_loss_set:
                pass
        with self.assertRaisesRegex(
            RP2ValueError, "Timestamp .* of acquired_lot entry .* is incompatible with timestamp .* of its ancestor .* using .* accounting method: .*"
        ):
            gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.2"), out15, in_transaction_test))
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("2"), self._in2[asset], None))
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.01"), out14, in3))
            for _ in gain_loss_set:
                pass
        with self.assertRaisesRegex(RP2ValueError, "Entry already added: GainLoss"):
            gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)
            gain_loss_set.add_entry(gain_loss1)
            gain_loss_set.add_entry(gain_loss1)
        with self.assertRaisesRegex(RP2ValueError, "Entry already added: GainLoss"):
            gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)
            gain_loss_set.add_entry(gain_loss1)
            # Different instance with same contents as gain_loss1
            gain_loss_test: GainLoss = GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.2"), self._out15[asset], self._in3[asset])
            gain_loss_set.add_entry(gain_loss_test)
        with self.assertRaisesRegex(RP2ValueError, "Taxable event crypto amount already exhausted for OutTransaction"):
            gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.2"), out15, in3))
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.2"), out15, in_transaction_test))
            for _ in gain_loss_set:
                pass
        with self.assertRaisesRegex(RP2ValueError, "Acquired lot crypto amount already exhausted for InTransaction"):
            gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("1"), out14, in3))
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("1"), out16, in3))
            for _ in gain_loss_set:
                pass
        with self.assertRaisesRegex(RP2ValueError, "Current taxable event amount .* exceeded crypto balance change of taxable event .* GainLoss"):
            gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("3"), out16, in6))
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("4"), out16, in5))
            for _ in gain_loss_set:
                pass
        with self.assertRaisesRegex(RP2ValueError, "Current acquired lot amount .* exceeded crypto balance change of acquired lot .* GainLoss"):
            gain_loss_set = GainLossSet(self._configuration, self._accounting_method, asset)
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.2"), out15, in3))
            gain_loss_set.add_entry(GainLoss(self._configuration, self._accounting_method, RP2Decimal("1"), out14, in3))
            for _ in gain_loss_set:
                pass


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from rp2.abstract_accounting_method import AbstractAccountingMethod
from rp2.computed_data import ComputedData
from rp2.configuration import Configuration
from rp2.input_data import InputData
from rp2.ods_parser import open_ods, parse_ods
from rp2.out_transaction import OutTransaction
from rp2.plugin.accounting_method.fifo import AccountingMethod
from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError
from rp2.tax_engine import compute_tax
from rp2_test_output import RP2_TEST_OUTPUT  # pylint: disable=wrong-import-order


class TestTaxEngine(unittest.TestCase):
    _good_input_configuration: Configuration
    _bad_input_configuration: Configuration
    _accounting_method: AbstractAccountingMethod

    @classmethod
    def setUpClass(cls) -> None:
        TestTaxEngine._good_input_configuration = Configuration("./config/test_data.config", US())
        TestTaxEngine._bad_input_configuration = Configuration("./config/test_bad_data.config", US())
        TestTaxEngine._accounting_method = AccountingMethod()

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_good_input(self) -> None:
        self._verify_good_output("B1")
        self._verify_good_output("B2")
        self._verify_good_output("B3")
        self._verify_good_output("B4")

    def _verify_good_output(self, sheet_name: str) -> None:
        asset = sheet_name

        # Parser is tested separately (on same input) in test_input_parser.py
        input_file_handle: object = open_ods(self._good_input_configuration, "./input/test_data.ods")
        input_data: InputData = parse_ods(self._good_input_configuration, asset, input_file_handle)

        # In table is always present
        computed_data: ComputedData = compute_tax(self._good_input_configuration, self._accounting_method, input_data)

        if asset in RP2_TEST_OUTPUT:
            self.assertEqual(str(computed_data.gain_loss_set), RP2_TEST_OUTPUT[asset])

    def test_bad_input(self) -> None:
        asset = "B4"
        input_file_handle: object = open_ods(self._good_input_configuration, "./input/test_data.ods")
        input_data: InputData = parse_ods(self._bad_input_configuration, asset, input_file_handle)

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            compute_tax(
                None,  # type: ignore
                self._accounting_method,
                input_data,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            compute_tax(
                1111,  # type: ignore
                self._accounting_method,
                input_data,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'accounting_method' is not of type AbstractAccountingMethod: .*"):
            compute_tax(
                self._good_input_configuration,
                None,  # type: ignore
                input_data,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'accounting_method' is not of type AbstractAccountingMethod: .*"):
            compute_tax(
                self._good_input_configuration,
                1111,  # type: ignore
                input_data,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'input_data' is not of type InputData: .*"):
            compute_tax(
                self._bad_input_configuration,
                self._accounting_method,
                None,  # type: ignore
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'input_data' is not of type InputData: .*"):
            compute_tax(
                self._bad_input_configuration,
                self._accounting_method,
                "foobar",  # type: ignore
            )

        input_data.unfiltered_out_transaction_set.add_entry(
            OutTransaction(
                self._good_input_configuration,
                "6/1/2020 3:59:59 -04:00",
                asset,
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("20.2"),
                RP2Decimal("1"),
                unique_id=38,
            )
        )

        with self.assertRaisesRegex(RP2ValueError, "Total in-transaction value < total taxable entries"):
            compute_tax(
                self._good_input_configuration,
                self._accounting_method,
                input_data,
            )


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Optional, Set

from rp2.configuration import Configuration
from rp2.rp2_error import RP2TypeError, RP2ValueError


class TransactionType(Enum):
    AIRDROP: str = "airdrop"
    BUY: str = "buy"
    DONATE: str = "donate"
    GIFT: str = "gift"
    HARDFORK: str = "hardfork"
    INTEREST: str = "interest"
    MINING: str = "mining"
    MOVE: str = "move"
    SELL: str = "sell"
    STAKING: str = "staking"
    WAGES: str = "wages"

    @classmethod
    def has_value(cls, value: str) -> bool:
        return value in _transaction_type_values

    @classmethod
    def type_check_from_string(cls, name: str, transaction_type: str) -> "TransactionType":
        Configuration.type_check_string(name, transaction_type)
        if not TransactionType.has_value(transaction_type.lower()):
            raise RP2ValueError(f"Parameter '{name}' has invalid transaction type value: {transaction_type}")
        return TransactionType[transaction_type.upper()]

    @classmethod
    def type_check(cls, name: str, transaction_type: "TransactionType") -> "TransactionType":
        Configuration.type_check_parameter_name(name)
        if not isinstance(transaction_type, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {transaction_type}")
        return transaction_type

    def is_earn_type(self) -> bool:
        return self in _transaction_type_earn_values


_transaction_type_values: Set[str] = {item.value for item in TransactionType}
_transaction_type_earn_values: Set[TransactionType] = {
    TransactionType.AIRDROP,
    TransactionType.HARDFORK,
    TransactionType.INTEREST,
    TransactionType.MINING,
    TransactionType.STAKING,
    TransactionType.WAGES,
}


class EntrySetType(Enum):
    IN: str = "in"
    INTRA: str = "intra"
    MIXED: str = "mixed"
    OUT: str = "out"

    @classmethod
    def has_value(cls, value: str) -> bool:
        return value in _entry_set_type_values

    @classmethod
    def get_entry_set_type_from_string(cls, entry_set_type: str) -> Optional["EntrySetType"]:
        if not isinstance(entry_set_type, str):
            # Disable mypy because otherwise it warns about unreachable code (we still want this runtime check)
            return None  # type: ignore
        if not EntrySetType.has_value(entry_set_type.lower()):
            return None
        return EntrySetType[entry_set_type.upper()]

    @classmethod
    def type_check_from_string(cls, name: str, entry_set_type: str) -> "EntrySetType":
        Configuration.type_check_string(name, entry_set_type)
        if not isinstance(entry_set_type, str):
            raise RP2TypeError(f"Parameter '{name}' is not of type string: {entry_set_type}")
        result = cls.get_entry_set_type_from_string(entry_set_type)
        if result is None:
            raise RP2ValueError(f"Parameter '{name}' has invalid entry set type value: {entry_set_type}")
        return result

    @classmethod
    def type_check(cls, name: str, entry_set_type: "EntrySetType") -> "EntrySetType":
        Configuration.type_check_parameter_name(name)
        if not isinstance(entry_set_type, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {entry_set_type}")
        return entry_set_type


_entry_set_type_values: Set[str] = {item.value for item in EntrySetType}



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import List, Optional

from rp2.configuration import Configuration, to_string
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError


class AbstractEntry:
    def __init__(
        self,
        configuration: Configuration,
        asset: str,
    ) -> None:
        self.__configuration = Configuration.type_check("configuration", configuration)
        self.__asset: str = configuration.type_check_asset("asset", asset)

    @classmethod
    def type_check(cls, name: str, instance: "AbstractEntry") -> "AbstractEntry":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:
        class_specific_data: List[str] = []
        if repr_format:
            class_specific_data.append(f"{type(self).__name__}(id={repr(self.unique_id)}")
        else:
            class_specific_data.append(f"{type(self).__name__}:")
            class_specific_data.append(f"id={str(self.unique_id)}")

        if extra_data:
            class_specific_data.extend(extra_data)

        return to_string(indent=indent, repr_format=repr_format, data=class_specific_data)

    def __str__(self) -> str:
        return self.to_string(indent=0, repr_format=False)

    def __repr__(self) -> str:
        return self.to_string(indent=0, repr_format=True)

    # Used for hashing but there are not enough attributes in this class, so the method is abstract.
    def __eq__(self, other: object) -> bool:
        raise NotImplementedError("Abstract method")

    # Used for hashing but there are not enough attributes in this class, so the method is abstract.
    def __ne__(self, other: object) -> bool:
        raise NotImplementedError("Abstract method")

    # Used for hashing but there are not enough attributes in this class, so the method is abstract.
    def __hash__(self) -> int:
        raise NotImplementedError("Abstract method")

    @property
    def configuration(self) -> Configuration:
        return self.__configuration

    @property
    def asset(self) -> str:
        return self.__asset

    @property
    def unique_id(self) -> str:
        raise NotImplementedError("Abstract property")

    @property
    def timestamp(self) -> datetime:
        raise NotImplementedError("Abstract property")

    # How much crypto was gained / lost with this entry
    @property
    def crypto_balance_change(self) -> RP2Decimal:
        raise NotImplementedError("Abstract property")

    # How much fiat was gained / lost with this entry
    @property
    def fiat_balance_change(self) -> RP2Decimal:
        raise NotImplementedError("Abstract property")



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date
import logging
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Set, cast

import ezodf
from rp2.abstract_country import AbstractCountry
from rp2.abstract_entry import AbstractEntry
from rp2.abstract_transaction import AbstractTransaction
from rp2.balance import BalanceSet
from rp2.computed_data import ComputedData, YearlyGainLoss
from rp2.gain_loss import GainLoss
from rp2.gain_loss_set import GainLossSet
from rp2.in_transaction import InTransaction
from rp2.intra_transaction import IntraTransaction
from rp2.logger import create_logger
from rp2.out_transaction import OutTransaction
from rp2.plugin.report.abstract_ods_generator import AbstractODSGenerator
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError
from rp2.transaction_set import TransactionSet


LOGGER: logging.Logger = create_logger("rp2_full_report")

class _TransactionVisualStyle(NamedTuple):
    year: int
    visual_style: str
    highlighted_style: str


class _BorderStyle(NamedTuple):
    year: int
    border_suffix: str


class _AssetAndYear(NamedTuple):
    asset: str
    year: int


_ZERO: RP2Decimal = RP2Decimal(0)


class Generator(AbstractODSGenerator):

    MIN_ROWS: int = 40
    MAX_COLUMNS: int = 40
    OUTPUT_FILE: str = "rp2_full_report.ods"

    TEMPLATE_SHEETS_TO_KEEP: Set[str] = {"__Summary"}

    __in_out_sheet_transaction_2_row: Dict[AbstractTransaction, int] = {}
    __tax_sheet_year_2_row: Dict[_AssetAndYear, int] = {}

    __in_header_names_row_1: List[str] = []
    __in_header_names_row_2: List[str] = []
    __out_header_names_row_1: List[str] = []
    __out_header_names_row_2: List[str] = []
    __intra_header_names_row_1: List[str] = []
    __intra_header_names_row_2: List[str] = []
    __balance_header_names_row_1: List[str] = []
    __balance_header_names_row_2: List[str] = []
    __gain_loss_summary_header_names_row_1: List[str] = []
    __gain_loss_summary_header_names_row_2: List[str] = []
    __gain_loss_detail_header_names_row_1: List[str] = []
    __gain_loss_detail_header_names_row_2: List[str] = []

    def _setup_header_rows(self, country: AbstractCountry) -> None:

        currency_code: str = country.currency_iso_code.upper()

        self.__in_header_names_row_1: List[str] = [
            "",
            "",
            "",
            "",
            "",
            "Transaction",
            "",
            "Crypto",
            "Crypto In",
            "",
            f"{currency_code} In",
            f"{currency_code} In",
            "Taxable",
            "",
            "",
        ]

        self.__in_header_names_row_2: List[str] = [
            "Sent/Sold",
            "Timestamp",
            "Asset",
            "Exchange",
            "Holder",
            "Type",
            "Spot Price",
            "In",
            "Running Sum",
            f"{currency_code} Fee",
            "No Fee",
            "With Fee",
            "Event",
            "N/A",
            "Notes",
        ]

        self.__out_header_names_row_1: List[str] = [
            "",
            "",
            "",
            "",
            "Transaction",
            "",
            "",
            "",
            "Crypto Out",
            "Crypto Fee",
            "",
            "",
            "Taxable",
            "",
        ]

        self.__out_header_names_row_2: List[str] = [
            "Timestamp",
            "Asset",
            "Exchange",
            "Holder",
            "Type",
            "Spot Price",
            "Crypto Out",
            "Crypto Fee",
            "Running Sum",
            "Running Sum",
            f"{currency_code} Out",
            f"{currency_code} Fee",
            "Event",
            "Notes",
        ]

        self.__intra_header_names_row_1: List[str] = [
            "",
            "",
            "From",
            "From",
            "",
            "",
            "",
            "",
            "Crypto",
            "",
            "Crypto Fee",
            "",
            "Taxable",
            "",
        ]

        self.__intra_header_names_row_2: List[str] = [
            "Timestamp",
            "Asset",
            "Exchange",
            "Holder",
            "To Exchange",
            "To Holder",
            "Spot Price",
            "Crypto Sent",
            "Received",
            "Crypto Fee",
            "Running Sum",
            f"{currency_code} Fee",
            "Event",
            "Notes",
        ]

        self.__balance_header_names_row_1: List[str] = [
            "",
            "",
            "",
            "Acquired",
            "Sent",
            "Received",
            "Final",
        ]

        self.__balance_header_names_row_2: List[str] = [
            "Exchange",
            "Holder",
            "Asset",
            "Balance",
            "Balance",
            "Balance",
            "Balance",
        ]

        self.__gain_loss_summary_header_names_row_1: List[str] = [
            "",
            "",
            "Capital",
            "Capital",
            "Transaction",
            "Crypto",
            f"{currency_code}",
            f"{currency_code} Total",
        ]

        self.__gain_loss_summary_header_names_row_2: List[str] = [
            "Year",
            "Asset",
            "Gains",
            "Gains Type",
            "Type",
            "Taxable Total",
            "Taxable Total",
            "Cost Basis",
        ]

        self.__gain_loss_detail_header_names_row_1: List[str] = [
            "Crypto",
            "",
            "Crypto Amt",
            "Capital",
            "Capital",
            "Taxable Event",
            "Taxable Event",
            "Taxable Event",
            f"Taxable Event {currency_code}",
            "Taxable Event",
            "Taxable Event",
            "Acquired Lot",
            "Acquired Lot",
            f"Acquired Lot {currency_code}",
            f"Acquired Lot {currency_code}",
            f"Acquired Lot {currency_code}",
            "Acquired Lot",
            "Acquired Lot Fraction",
        ]

        self.__gain_loss_detail_header_names_row_2: List[str] = [
            "Amount",
            "Asset",
            "Running Sum",
            "Gains",
            "Gains Type",
            "Timestamp",
            "Direction/Type",
            "Fraction %",
            "Amount Fraction",
            "Spot Price",
            "Fraction Description",
            "Timestamp",
            "Fraction %",
            "Amount Fraction",
            "Fee Fraction",
            "Cost Basis",
            "Spot Price",
            "Description",
        ]

    def generate(
        self,
        country: AbstractCountry,
        accounting_method: str,
        asset_to_computed_data: Dict[str, ComputedData],
        output_dir_path: str,
        output_file_prefix: str,
        from_date: date,
        to_date: date,
    ) -> None:

        if not isinstance(asset_to_computed_data, Dict):
            raise RP2TypeError(f"Parameter 'asset_to_computed_data' has non-Dict value {asset_to_computed_data}")

        self._setup_header_rows(country)

        output_file: Any
        output_file = self._initialize_output_file(
            country=country,
            accounting_method=accounting_method,
            output_dir_path=output_dir_path,
            output_file_prefix=output_file_prefix,
            output_file_name=self.OUTPUT_FILE,
            template_sheets_to_keep=self.TEMPLATE_SHEETS_TO_KEEP,
            from_date=from_date,
            to_date=to_date,
        )

        asset: str
        computed_data: ComputedData

        summary_row_index: int = 3
        for asset, computed_data in asset_to_computed_data.items():
            if not isinstance(asset, str):
                raise RP2TypeError(f"Parameter 'asset' has non-string value {asset}")
            ComputedData.type_check("computed_data", computed_data)
            summary_row_index = self.__generate_asset(computed_data, output_file, summary_row_index)

        output_file.save()
        LOGGER.info("Plugin '%s' output: %s", __name__, Path(output_file.docname).resolve())

    @staticmethod
    def get_in_out_sheet_name(asset: str) -> str:
        return f"{asset} In-Out"

    @staticmethod
    def get_tax_sheet_name(asset: str) -> str:
        return f"{asset} Tax"

    def __get_number_of_rows_in_transaction_sheet(self, computed_data: ComputedData) -> int:
        return self.MIN_ROWS + computed_data.in_transaction_set.count + computed_data.out_transaction_set.count + computed_data.intra_transaction_set.count

    def __get_number_of_rows_in_output_sheet(self, computed_data: ComputedData) -> int:
        return self.MIN_ROWS + len(computed_data.yearly_gain_loss_list) + computed_data.balance_set.count + computed_data.gain_loss_set.count

    def __generate_asset(self, computed_data: ComputedData, output_file: Any, summary_row_index: int) -> int:
        asset: str = computed_data.asset
        transaction_sheet_name: str = self.get_in_out_sheet_name(asset)
        output_sheet_name: str = self.get_tax_sheet_name(asset)

        transaction_sheet: Any = ezodf.Table(transaction_sheet_name)
        output_sheet: Any = ezodf.Table(output_sheet_name)
        summary_sheet: Any = output_file.sheets["Summary"]

        output_file.sheets += transaction_sheet
        output_file.sheets += output_sheet

        transaction_sheet.reset(size=(self.__get_number_of_rows_in_transaction_sheet(computed_data), self.MAX_COLUMNS))
        output_sheet.reset(size=(self.__get_number_of_rows_in_output_sheet(computed_data), self.MAX_COLUMNS))

        new_lines: int = len(computed_data.yearly_gain_loss_list)
        if new_lines:
            summary_sheet.append_rows(new_lines)

        row_index: int = 0
        row_index = self.__generate_in_table(transaction_sheet, computed_data, row_index)
        row_index = self.__generate_out_table(transaction_sheet, computed_data, row_index + 2)
        row_index = self.__generate_intra_table(transaction_sheet, computed_data, row_index + 2)

        row_index = 0
        row_index = self.__generate_gain_loss_summary(output_sheet, computed_data.yearly_gain_loss_list, row_index)
        row_index = self.__generate_account_balances(output_sheet, computed_data.balance_set, row_index + 2)
        row_index = self.__generate_average_price_per_unit(output_sheet, asset, computed_data.price_per_unit, row_index + 2)
        row_index = self.__generate_gain_loss_detail(output_sheet, asset, computed_data, row_index + 2)

        return self.__generate_yearly_gain_loss_summary(summary_sheet, asset, computed_data.yearly_gain_loss_list, summary_row_index)

    @staticmethod
    def __get_transaction_visual_style(transaction: AbstractTransaction, year: int) -> _TransactionVisualStyle:
        visual_style: str = "transparent"
        highlighted_style: str = "transparent"
        if transaction.is_taxable():
            visual_style = "taxable_event"
            highlighted_style = "highlighted"
        if year == 0:
            year = transaction.timestamp.year
        if transaction.timestamp.year != year:
            visual_style = f"{visual_style}_border"
            highlighted_style = f"{highlighted_style}_border"
            year = transaction.timestamp.year
        return _TransactionVisualStyle(year, visual_style, highlighted_style)

    @staticmethod
    def __get_border_style(current_year: int, year: int) -> _BorderStyle:
        border_suffix: str = ""
        if year == 0:
            year = current_year
        if current_year != year:
            border_suffix = "_border"
            year = current_year
        return _BorderStyle(year, border_suffix)

    def __generate_in_table(self, sheet: Any, computed_data: ComputedData, row_index: int) -> int:
        row_index = self._fill_header("In-Flow Detail", self.__in_header_names_row_1, self.__in_header_names_row_2, sheet, row_index, 0)

        in_transaction_set: TransactionSet = computed_data.in_transaction_set
        entry: AbstractEntry
        year: int = 0
        visual_style: str
        previous_transaction: Optional[InTransaction] = None
        border_style: _BorderStyle
        border_suffix: str = ""
        for entry in in_transaction_set:
            transaction: InTransaction = cast(InTransaction, entry)
            highlighted_style: str
            transaction_visual_style: _TransactionVisualStyle = self.__get_transaction_visual_style(transaction, year)
            year = transaction_visual_style.year
            border_style = self.__get_border_style(entry.timestamp.year, year)
            border_suffix = border_style.border_suffix
            visual_style = transaction_visual_style.visual_style
            highlighted_style = transaction_visual_style.highlighted_style
            in_lot_sold_percentage: Optional[RP2Decimal] = computed_data.get_in_lot_sold_percentage(transaction)

            # Write _ZERO only on the first in_transaction if there are no sold lots
            if in_lot_sold_percentage == _ZERO and previous_transaction is not None:
                in_lot_sold_percentage = None
            self._fill_cell(
                sheet,
                row_index,
                0,
                in_lot_sold_percentage if in_lot_sold_percentage is not None else "",
                data_style="percent",
                visual_style="acquired_lot" + border_suffix if in_lot_sold_percentage is not None else "transparent",
            )
            self._fill_cell(sheet, row_index, 1, transaction.timestamp, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, transaction.asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 3, transaction.exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, transaction.holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, transaction.transaction_type.value.upper(), visual_style=visual_style)
            self._fill_cell(sheet, row_index, 6, transaction.spot_price, data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 7, transaction.crypto_in, data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 8, computed_data.get_crypto_in_running_sum(transaction), data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 9, transaction.fiat_fee, data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 10, transaction.fiat_in_no_fee, data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 11, transaction.fiat_in_with_fee, data_style="fiat", visual_style=highlighted_style)
            self._fill_cell(sheet, row_index, 12, "YES" if transaction.is_taxable() else "NO", data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 13, "", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 14, transaction.notes, visual_style="transparent")

            self.__in_out_sheet_transaction_2_row[transaction] = row_index + 1

            previous_transaction = transaction
            row_index += 1

        return row_index

    def __generate_out_table(self, sheet: Any, computed_data: ComputedData, row_index: int) -> int:
        row_index = self._fill_header("Out-Flow Detail", self.__out_header_names_row_1, self.__out_header_names_row_2, sheet, row_index, 1)

        out_transaction_set: TransactionSet = computed_data.out_transaction_set

        entry: AbstractEntry
        year: int = 0
        for entry in out_transaction_set:
            transaction: OutTransaction = cast(OutTransaction, entry)
            visual_style: str
            highlighted_style: str
            transaction_visual_style: _TransactionVisualStyle = self.__get_transaction_visual_style(transaction, year)
            year = transaction_visual_style.year
            visual_style = transaction_visual_style.visual_style
            highlighted_style = transaction_visual_style.highlighted_style
            self._fill_cell(sheet, row_index, 0, "", visual_style="transparent")
            self._fill_cell(sheet, row_index, 1, transaction.timestamp, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, transaction.asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 3, transaction.exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, transaction.holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, transaction.transaction_type.value.upper(), visual_style=visual_style)
            self._fill_cell(sheet, row_index, 6, transaction.spot_price, visual_style=visual_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 7, transaction.crypto_out_no_fee, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 8, transaction.crypto_fee, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 9, computed_data.get_crypto_out_running_sum(transaction), data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 10, computed_data.get_crypto_out_fee_running_sum(transaction), data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 11, transaction.crypto_out_no_fee * transaction.spot_price, visual_style=highlighted_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 12, transaction.crypto_fee * transaction.spot_price, visual_style=visual_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 13, "YES" if transaction.is_taxable() else "NO", data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 14, transaction.notes, visual_style="transparent")

            self.__in_out_sheet_transaction_2_row[transaction] = row_index + 1

            row_index += 1

        return row_index

    def __generate_intra_table(self, sheet: Any, computed_data: ComputedData, row_index: int) -> int:
        row_index = self._fill_header("Intra-Flow Detail", self.__intra_header_names_row_1, self.__intra_header_names_row_2, sheet, row_index, 1)

        intra_transaction_set: TransactionSet = computed_data.intra_transaction_set

        entry: AbstractEntry
        year: int = 0
        for entry in intra_transaction_set:
            transaction: IntraTransaction = cast(IntraTransaction, entry)
            visual_style: str
            highlighted_style: str
            transaction_visual_style: _TransactionVisualStyle = self.__get_transaction_visual_style(transaction, year)
            year = transaction_visual_style.year
            visual_style = transaction_visual_style.visual_style
            highlighted_style = transaction_visual_style.highlighted_style
            self._fill_cell(sheet, row_index, 0, "", visual_style="transparent")
            self._fill_cell(sheet, row_index, 1, transaction.timestamp, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, transaction.asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 3, transaction.from_exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, transaction.from_holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, transaction.to_exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 6, transaction.to_holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 7, transaction.spot_price, visual_style=visual_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 8, transaction.crypto_sent, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 9, transaction.crypto_received, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 10, transaction.crypto_fee, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 11, computed_data.get_crypto_intra_fee_running_sum(transaction), data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 12, transaction.fiat_fee, visual_style=highlighted_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 13, "YES" if transaction.is_taxable() else "NO", data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 14, transaction.notes, visual_style="transparent")

            self.__in_out_sheet_transaction_2_row[transaction] = row_index + 1

            row_index += 1

        return row_index

    def __generate_gain_loss_summary(self, sheet: Any, yearly_gain_loss_list: List[YearlyGainLoss], row_index: int) -> int:
        row_index = self._fill_header(
            "Gain / Loss Summary", self.__gain_loss_summary_header_names_row_1, self.__gain_loss_summary_header_names_row_2, sheet, row_index, 0
        )

        year: int = 0
        for yearly_gain_loss in yearly_gain_loss_list:
            border_suffix: str = ""
            capital_gains_type: str = "LONG" if yearly_gain_loss.is_long_term_capital_gains else "SHORT"
            border_style: _BorderStyle = self.__get_border_style(yearly_gain_loss.year, year)
            year = border_style.year
            border_suffix = border_style.border_suffix
            self._fill_cell(sheet, row_index, 0, yearly_gain_loss.year, visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 1, yearly_gain_loss.asset, visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 2, yearly_gain_loss.fiat_gain_loss, visual_style="bold" + border_suffix, data_style="fiat")
            self._fill_cell(sheet, row_index, 3, capital_gains_type, visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 4, yearly_gain_loss.transaction_type.value.upper(), visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 5, yearly_gain_loss.crypto_amount, visual_style="transparent" + border_suffix, data_style="crypto")
            self._fill_cell(sheet, row_index, 6, yearly_gain_loss.fiat_amount, visual_style="taxable_event" + border_suffix, data_style="fiat")
            self._fill_cell(sheet, row_index, 7, yearly_gain_loss.fiat_cost_basis, visual_style="acquired_lot" + border_suffix, data_style="fiat")
            row_index += 1

        return row_index

    def __generate_account_balances(self, sheet: Any, balance_set: BalanceSet, row_index: int) -> int:
        row_index = self._fill_header("Account Balances", self.__balance_header_names_row_1, self.__balance_header_names_row_2, sheet, row_index, 0)

        totals: Dict[str, RP2Decimal] = {}
        value: RP2Decimal
        for balance in balance_set:
            self._fill_cell(sheet, row_index, 0, balance.exchange, visual_style="bold", data_style="default")
            self._fill_cell(sheet, row_index, 1, balance.holder, visual_style="bold", data_style="default")
            self._fill_cell(sheet, row_index, 2, balance.asset)
            self._fill_cell(sheet, row_index, 3, balance.acquired_balance, data_style="crypto")
            self._fill_cell(sheet, row_index, 4, balance.sent_balance, data_style="crypto")
            self._fill_cell(sheet, row_index, 5, balance.received_balance, data_style="crypto")
            self._fill_cell(sheet, row_index, 6, balance.final_balance, visual_style="bold", data_style="crypto")
            value = totals.setdefault(balance.holder, _ZERO)
            value += balance.final_balance
            totals[balance.holder] = value
            row_index += 1

        holder: str
        border_drawn: bool = False
        for (holder, value) in sorted(totals.items()):
            border_suffix: str = ""
            if not border_drawn:
                border_suffix = "_border"
                border_drawn = True
            self._fill_cell(sheet, row_index, 0, "Total", visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 1, holder, visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 2, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 3, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 4, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 5, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 6, value, visual_style="bold" + border_suffix, data_style="crypto")
            row_index += 1

        return row_index

    def __generate_average_price_per_unit(self, sheet: Any, asset: str, price_per_unit: RP2Decimal, row_index: int) -> int:
        self._fill_cell(sheet, row_index, 0, "Average Price", visual_style="title")
        self._fill_cell(sheet, row_index + 1, 0, "Average Price", visual_style="header")
        self._fill_cell(sheet, row_index + 2, 0, f"Paid Per 1 {asset}", visual_style="header")
        self._fill_cell(sheet, row_index + 3, 0, price_per_unit, visual_style="transparent", data_style="fiat")

        return row_index + 4

    def __get_hyperlinked_transaction_value(self, transaction: AbstractTransaction, value: Any) -> Any:
        row: Optional[int] = self.__get_in_out_sheet_row(transaction)
        if not row:
            # This may occur if command line time filters are activated
            return value
        if isinstance(value, (RP2Decimal, int, float)):
            return f'=HYPERLINK("#{self.get_in_out_sheet_name(transaction.asset)}.a{row}:z{row}"; {value})'
        return f'=HYPERLINK("#{self.get_in_out_sheet_name(transaction.asset)}.a{row}:z{row}"; "{value}")'

    def __get_hyperlinked_summary_value(self, asset: str, value: Any, year: int) -> Any:
        row: int = self.__tax_sheet_year_2_row[_AssetAndYear(asset, year)]
        if isinstance(value, (RP2Decimal, int, float)):
            return f'=HYPERLINK("#{self.get_tax_sheet_name(asset)}.a{row}:z{row}"; {value})'
        return f'=HYPERLINK("#{self.get_tax_sheet_name(asset)}.a{row}:z{row}"; "{value}")'

    def __get_in_out_sheet_row(self, transaction: AbstractTransaction) -> Optional[int]:
        if transaction not in self.__in_out_sheet_transaction_2_row:
            return None
        return self.__in_out_sheet_transaction_2_row[transaction]

    def __generate_gain_loss_detail(self, sheet: Any, asset: str, computed_data: ComputedData, row_index: int) -> int:

        row_index = self._fill_header(
            "Gain / Loss Detail", self.__gain_loss_detail_header_names_row_1, self.__gain_loss_detail_header_names_row_2, sheet, row_index, 0
        )

        gain_loss_set: GainLossSet = computed_data.gain_loss_set

        taxable_event_style_modifier: str = ""
        acquired_lot_style_modifier: str = "_alt"
        year: int = 0
        border_style: _BorderStyle

        previous_acquired_lot: Optional[InTransaction] = None
        for entry in gain_loss_set:
            gain_loss: GainLoss = cast(GainLoss, entry)
            border_suffix: str = ""
            border_style = self.__get_border_style(gain_loss.taxable_event.timestamp.year, year)
            if gain_loss.taxable_event.timestamp.year != year:
                self.__tax_sheet_year_2_row[_AssetAndYear(asset, gain_loss.taxable_event.timestamp.year)] = row_index + 1
            year = border_style.year
            border_suffix = border_style.border_suffix
            transparent_style: str = f"transparent{border_suffix}"
            taxable_event_style: str = f"taxable_event{taxable_event_style_modifier}{border_suffix}"
            highlighted_style: str = f"highlighted{border_suffix}"
            current_taxable_event_fraction: int = gain_loss_set.get_taxable_event_fraction(gain_loss) + 1
            total_taxable_event_fractions: int = gain_loss_set.get_taxable_event_number_of_fractions(gain_loss.taxable_event)
            transaction_type: str = (
                f"{self._get_table_type_from_transaction(gain_loss.taxable_event)} / " f"{gain_loss.taxable_event.transaction_type.value.upper()}"
            )
            taxable_event_note: str = (
                f"{current_taxable_event_fraction}/"
                f"{total_taxable_event_fractions}: "
                f"{gain_loss.crypto_amount:.8f} of "
                f"{gain_loss.taxable_event.crypto_balance_change:.8f} "
                f"{asset}"
            )
            acquired_lot_style: str

            self._fill_cell(sheet, row_index, 0, gain_loss.crypto_amount, visual_style=transparent_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 1, gain_loss.asset, visual_style=transparent_style)
            self._fill_cell(sheet, row_index, 2, computed_data.get_crypto_gain_loss_running_sum(gain_loss), visual_style=transparent_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 3, gain_loss.fiat_gain, visual_style=transparent_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 4, "LONG" if gain_loss.is_long_term_capital_gains() else "SHORT", visual_style=transparent_style)
            self._fill_cell(
                sheet,
                row_index,
                5,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, gain_loss.taxable_event.timestamp),
                visual_style=taxable_event_style,
            )
            self._fill_cell(
                sheet, row_index, 6, self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, transaction_type), visual_style=taxable_event_style
            )
            self._fill_cell(
                sheet,
                row_index,
                7,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, gain_loss.taxable_event_fraction_percentage),
                visual_style=taxable_event_style,
                data_style="percent",
            )
            self._fill_cell(
                sheet,
                row_index,
                8,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, gain_loss.taxable_event_fiat_amount_with_fee_fraction),
                visual_style=highlighted_style,
                data_style="fiat",
            )
            self._fill_cell(
                sheet,
                row_index,
                9,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, gain_loss.taxable_event.spot_price),
                visual_style=taxable_event_style,
                data_style="fiat",
            )
            self._fill_cell(
                sheet,
                row_index,
                10,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, taxable_event_note),
                visual_style=f"taxable_event_note{border_suffix}",
            )
            if current_taxable_event_fraction == total_taxable_event_fractions:
                # Last fraction: change color
                taxable_event_style_modifier = "" if taxable_event_style_modifier == "_alt" else "_alt"

            if gain_loss.acquired_lot:
                if gain_loss.acquired_lot != previous_acquired_lot:
                    # Last fraction: change color
                    acquired_lot_style_modifier = "" if acquired_lot_style_modifier == "_alt" else "_alt"
                acquired_lot_style = f"acquired_lot{acquired_lot_style_modifier}{border_suffix}"
                current_acquired_lot_fraction: int = gain_loss_set.get_acquired_lot_fraction(gain_loss) + 1
                total_acquired_lot_fractions: int = gain_loss_set.get_acquired_lot_number_of_fractions(gain_loss.acquired_lot)
                acquired_lot_note: str = (
                    f"{current_acquired_lot_fraction}/"
                    f"{total_acquired_lot_fractions}: "
                    f"{gain_loss.crypto_amount:.8f} of "
                    f"{gain_loss.acquired_lot.crypto_balance_change:.8f} "
                    f"{asset}"
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    11,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.acquired_lot.timestamp),
                    visual_style=acquired_lot_style,
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    12,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.acquired_lot_fraction_percentage),
                    visual_style=acquired_lot_style,
                    data_style="percent",
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    13,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.acquired_lot_fiat_amount_with_fee_fraction),
                    visual_style=acquired_lot_style,
                    data_style="fiat",
                )
                fiat_fee_fraction: RP2Decimal = gain_loss.acquired_lot.fiat_fee * gain_loss.acquired_lot_fraction_percentage
                self._fill_cell(
                    sheet,
                    row_index,
                    14,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, fiat_fee_fraction),
                    visual_style=acquired_lot_style,
                    data_style="fiat",
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    15,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.fiat_cost_basis),
                    visual_style=highlighted_style,
                    data_style="fiat",
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    16,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.acquired_lot.spot_price),
                    visual_style=acquired_lot_style,
                    data_style="fiat",
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    17,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, acquired_lot_note),
                    visual_style=f"acquired_lot_note{border_suffix}",
                )

                previous_acquired_lot = gain_loss.acquired_lot
            else:
                acquired_lot_style = f"acquired_lot{acquired_lot_style_modifier}{border_suffix}"
                for i in range(11, 17):
                    self._fill_cell(sheet, row_index, i, "", visual_style=f"{acquired_lot_style}")

            row_index += 1

        return row_index

    def __generate_yearly_gain_loss_summary(self, sheet: Any, asset: str, yearly_gain_loss_list: List[YearlyGainLoss], row_index: int) -> int:
        for gain_loss in yearly_gain_loss_list:
            visual_style: str = "transparent"
            capital_gains_type: str = "LONG" if gain_loss.is_long_term_capital_gains else "SHORT"
            year: int = gain_loss.year
            self._fill_cell(sheet, row_index, 0, self.__get_hyperlinked_summary_value(asset, year, year), visual_style=visual_style)
            self._fill_cell(sheet, row_index, 1, self.__get_hyperlinked_summary_value(asset, asset, year), visual_style=visual_style)
            self._fill_cell(
                sheet, row_index, 2, self.__get_hyperlinked_summary_value(asset, gain_loss.fiat_gain_loss, year), visual_style=visual_style, data_style="fiat"
            )
            self._fill_cell(sheet, row_index, 3, self.__get_hyperlinked_summary_value(asset, capital_gains_type, year), visual_style=visual_style)
            self._fill_cell(
                sheet, row_index, 4, self.__get_hyperlinked_summary_value(asset, gain_loss.transaction_type.value.upper(), year), visual_style=visual_style
            )
            self._fill_cell(
                sheet, row_index, 5, self.__get_hyperlinked_summary_value(asset, gain_loss.crypto_amount, year), visual_style=visual_style, data_style="crypto"
            )
            self._fill_cell(
                sheet, row_index, 6, self.__get_hyperlinked_summary_value(asset, gain_loss.fiat_amount, year), visual_style=visual_style, data_style="fiat"
            )
            self._fill_cell(
                sheet, row_index, 7, self.__get_hyperlinked_summary_value(asset, gain_loss.fiat_cost_basis, year), visual_style=visual_style, data_style="fiat"
            )
            row_index += 1

        return row_index


def main() -> None:
    pass


if __name__ == "__main__":
    main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from datetime import date
from typing import Callable, Dict, List, Optional, cast

from rp2.configuration import Configuration, to_string
from rp2.in_transaction import InTransaction
from rp2.input_data import InputData
from rp2.intra_transaction import IntraTransaction
from rp2.logger import LOGGER
from rp2.out_transaction import OutTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError


@dataclass(frozen=True, eq=True)
class Balance:
    configuration: Configuration
    asset: str
    exchange: str
    holder: str
    final_balance: RP2Decimal
    acquired_balance: RP2Decimal
    sent_balance: RP2Decimal
    received_balance: RP2Decimal

    def __post_init__(self) -> None:
        Configuration.type_check("configuration", self.configuration)
        self.configuration.type_check_asset("asset", self.asset)
        self.configuration.type_check_exchange("exchange", self.exchange)
        self.configuration.type_check_holder("holder", self.holder)
        self.configuration.type_check_decimal("final_balance", self.final_balance)
        self.configuration.type_check_decimal("acquired_balance", self.acquired_balance)
        self.configuration.type_check_decimal("sent_balance", self.sent_balance)
        self.configuration.type_check_decimal("received_balance", self.received_balance)

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:
        class_specific_data: List[str] = []
        stringify: Callable[[object], str] = repr
        if not repr_format:
            stringify = str

        if repr_format:
            class_specific_data.append(f"{type(self).__name__}(asset={repr(self.asset)}")
        else:
            class_specific_data.append(f"{type(self).__name__}:")
            class_specific_data.append(f"asset={str(self.asset)}")

        class_specific_data.append(f"exchange={stringify(self.exchange)}")
        class_specific_data.append(f"holder={stringify(self.holder)}")
        class_specific_data.append(f"final_balance={self.final_balance:.8f}")
        class_specific_data.append(f"acquired_balance={self.acquired_balance:.8f}")
        class_specific_data.append(f"sent_balance={self.sent_balance:.8f}")
        class_specific_data.append(f"received_balance={self.received_balance:.8f}")

        if extra_data:
            class_specific_data.extend(extra_data)

        return to_string(indent=indent, repr_format=repr_format, data=class_specific_data)

    def __str__(self) -> str:
        return self.to_string(indent=0, repr_format=False)

    def __repr__(self) -> str:
        return self.to_string(indent=0, repr_format=True)


@dataclass(frozen=True, eq=True)
class Account:
    exchange: str
    holder: str


class BalanceSet:
    @classmethod
    def type_check(cls, name: str, instance: "BalanceSet") -> "BalanceSet":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    # from_date is not used when computing average price per unit: only to_date is relevant.
    def __init__(
        self,
        configuration: Configuration,
        input_data: InputData,
        to_date: date,
    ) -> None:

        Configuration.type_check("configuration", configuration)
        self.__input_data = InputData.type_check("input_data", input_data)

        self.__asset: str = configuration.type_check_asset("in_transaction_set.asset", input_data.asset)
        self._balances: List[Balance] = []

        acquired_balances: Dict[Account, RP2Decimal] = {}
        sent_balances: Dict[Account, RP2Decimal] = {}
        received_balances: Dict[Account, RP2Decimal] = {}
        final_balances: Dict[Account, RP2Decimal] = {}

        from_account: Account
        to_account: Account

        # Balances for bought and earned currency
        for transaction in self.__input_data.unfiltered_in_transaction_set:
            if transaction.timestamp.date() > to_date:
                break
            in_transaction: InTransaction = cast(InTransaction, transaction)
            to_account = Account(in_transaction.exchange, in_transaction.holder)
            acquired_balances[to_account] = acquired_balances.get(to_account, ZERO) + in_transaction.crypto_in
            final_balances[to_account] = final_balances.get(to_account, ZERO) + in_transaction.crypto_in

        # Balances for currency that is moved across accounts
        for transaction in self.__input_data.unfiltered_intra_transaction_set:
            if transaction.timestamp.date() > to_date:
                break
            intra_transaction: IntraTransaction = cast(IntraTransaction, transaction)
            from_account = Account(intra_transaction.from_exchange, intra_transaction.from_holder)
            to_account = Account(intra_transaction.to_exchange, intra_transaction.to_holder)
            sent_balances[from_account] = sent_balances.get(from_account, ZERO) + intra_transaction.crypto_sent
            received_balances[to_account] = received_balances.get(to_account, ZERO) + intra_transaction.crypto_received
            final_balances[from_account] = final_balances.get(from_account, ZERO) - intra_transaction.crypto_sent
            final_balances[to_account] = final_balances.get(to_account, ZERO) + intra_transaction.crypto_received

        # Balances for sold and gifted currency
        for transaction in self.__input_data.unfiltered_out_transaction_set:
            if transaction.timestamp.date() > to_date:
                break
            out_transaction: OutTransaction = cast(OutTransaction, transaction)
            from_account = Account(out_transaction.exchange, out_transaction.holder)
            sent_balances[from_account] = sent_balances.get(from_account, ZERO) + out_transaction.crypto_out_no_fee + out_transaction.crypto_fee
            final_balances[from_account] = final_balances.get(from_account, ZERO) - out_transaction.crypto_out_no_fee - out_transaction.crypto_fee

        for account, final_balance in final_balances.items():
            balance = Balance(
                configuration,
                self.__asset,
                account.exchange,
                account.holder,
                final_balance,
                acquired_balances.get(account, ZERO),
                sent_balances.get(account, ZERO),
                received_balances.get(account, ZERO),
            )
            LOGGER.debug("created balance: %s", balance)
            self._balances.append(balance)

        self._balances.sort(key=_balance_sort_key)

    def __str__(self) -> str:
        output: List[str] = []
        output.append(f"{type(self).__name__}:")
        output.append(f"  asset={self.asset}")
        output.append("  balances=")
        for balance in self:
            output.append(f"{balance.to_string(indent=2, repr_format=False)}")
        return "\n".join(output)

    def __repr__(self) -> str:
        output: List[str] = []
        output.append(f"{type(self).__name__}(")
        output.append(f", asset={repr(self.asset)}")
        output.append(", balances=[")
        count: int = 0
        for balance in self:
            if count > 0:
                output.append(", ")
            output.append(repr(balance))
            count += 1
        output.append("]")
        output.append(")")
        return "".join(output)

    def __iter__(self) -> "BalanceSetIterator":
        return BalanceSetIterator(self)

    @property
    def count(self) -> int:
        return len(self._balances)

    @property
    def asset(self) -> str:
        return self.__asset


class BalanceSetIterator:
    def __init__(self, balance_set: BalanceSet) -> None:
        self.__balance_set: BalanceSet = balance_set
        self.__balance_set_size: int = len(self.__balance_set._balances)
        self.__index: int = 0

    def __next__(self) -> Balance:
        result: Optional[Balance] = None
        if self.__index < self.__balance_set_size:
            result = self.__balance_set._balances[self.__index]  # pylint: disable=protected-access
            self.__index += 1
            return result
        raise StopIteration(self)


def _balance_sort_key(balance: Balance) -> str:
    return f"{balance.exchange}_{balance.holder}"



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional


class RP2Error(Exception):
    def __init__(self, message: str, unique_id: Optional[int] = None) -> None:
        self.__message = message if unique_id is None else f"{unique_id}: {message}"
        self.__unique_id = unique_id
        super().__init__(self.__message)

    def __repr__(self) -> str:
        return self.message

    @property
    def message(self) -> str:
        return self.__message

    @property
    def unique_id(self) -> Optional[int]:
        return self.__unique_id


class RP2TypeError(RP2Error):
    pass


class RP2ValueError(RP2Error):
    pass



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rp2.abstract_entry import AbstractEntry
from rp2.abstract_entry_set import AbstractEntrySet
from rp2.abstract_transaction import AbstractTransaction
from rp2.configuration import Configuration
from rp2.entry_types import EntrySetType
from rp2.rp2_error import RP2TypeError, RP2ValueError


class TransactionSet(AbstractEntrySet):
    @classmethod
    def type_check(cls, name: str, instance: "TransactionSet", entry_set_type: EntrySetType, asset: str, allow_empty: bool = False) -> "TransactionSet":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter is not of type {cls.__name__}: {instance}")
        if instance.entry_set_type != entry_set_type:
            raise RP2ValueError(f"entry_set_type {instance.entry_set_type} != {entry_set_type}: {instance}")
        if instance.asset != asset:
            raise RP2ValueError(f"asset {instance.asset} != {asset}: {instance}")
        if not allow_empty and instance.is_empty():
            raise RP2ValueError(f"IN transaction set is empty: {instance}")
        return instance

    def add_entry(self, entry: AbstractEntry) -> None:
        AbstractTransaction.type_check("entry", entry)
        super().add_entry(entry)

    def _validate_entry(self, entry: AbstractEntry) -> None:
        AbstractTransaction.type_check("entry", entry)
        super()._validate_entry(entry)



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-lines
RP2_TEST_OUTPUT = {
    "B1": """GainLossSet:
  configuration=./config/test_data.config
  asset=B1
  from_date=non-specified
  to_date=non-specified
  entries=
    GainLoss:
      id=3->None
      crypto_amount=2.00000000
      fiat_cost_basis=0.0000
      fiat_gain=24000.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=24000.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=InTransaction:
        id=3
        timestamp=2020-02-01 11:18:00.000000 +0000
        asset=B1
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=12000.0000
        crypto_in=2.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=24000.0000
        fiat_in_with_fee=24000.0000
        is_taxable=True
        fiat_taxable_amount=24000.0000
      acquired_lot_fiat_amount_with_fee_fraction=0.0000
      acquired_lot_fraction_percentage=0.0000%
      acquired_lot=None
      taxable_event_fraction=1 of 1
      parent=None
    GainLoss:
      id=7->None
      crypto_amount=3.00000000
      fiat_cost_basis=0.0000
      fiat_gain=39000.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=39000.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=InTransaction:
        id=7
        timestamp=2020-03-01 09:45:00.000000 +0000
        asset=B1
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=13000.0000
        crypto_in=3.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=39000.0000
        fiat_in_with_fee=39000.0000
        is_taxable=True
        fiat_taxable_amount=39000.0000
      acquired_lot_fiat_amount_with_fee_fraction=0.0000
      acquired_lot_fraction_percentage=0.0000%
      acquired_lot=None
      taxable_event_fraction=1 of 1
      parent=3->None""",
    "B2": """GainLossSet:
  configuration=./config/test_data.config
  asset=B2
  from_date=non-specified
  to_date=non-specified
  entries=
    GainLoss:
      id=16->4
      crypto_amount=0.20000000
      fiat_cost_basis=2220.0000
      fiat_gain=20.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=2240.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=OutTransaction:
        id=16
        timestamp=2020-01-11 11:15:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.SELL
        spot_price=11200.0000
        crypto_out_no_fee=0.20000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=2240.0000
      acquired_lot_fiat_amount_with_fee_fraction=2220.0000
      acquired_lot_fraction_percentage=20.0000%
      acquired_lot=InTransaction:
        id=4
        timestamp=2020-01-01 08:41:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=11000.0000
        crypto_in=1.00000000
        fiat_fee=100.0000
        fiat_in_no_fee=11000.0000
        fiat_in_with_fee=11100.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=1 of 2
      parent=None
    GainLoss:
      id=3->None
      crypto_amount=2.00000000
      fiat_cost_basis=0.0000
      fiat_gain=24000.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=24000.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=InTransaction:
        id=3
        timestamp=2020-02-01 11:18:00.000000 +0000
        asset=B2
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=12000.0000
        crypto_in=2.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=24000.0000
        fiat_in_with_fee=24000.0000
        is_taxable=True
        fiat_taxable_amount=24000.0000
      acquired_lot_fiat_amount_with_fee_fraction=0.0000
      acquired_lot_fraction_percentage=0.0000%
      acquired_lot=None
      taxable_event_fraction=1 of 1
      parent=16->4
    GainLoss:
      id=15->4
      crypto_amount=0.80000000
      fiat_cost_basis=8880.0000
      fiat_gain=880.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=9760.0000
      taxable_event_fraction_percentage=80.0000%
      taxable_event=OutTransaction:
        id=15
        timestamp=2020-02-11 19:58:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.SELL
        spot_price=12200.0000
        crypto_out_no_fee=1.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=12200.0000
      acquired_lot_fiat_amount_with_fee_fraction=8880.0000
      acquired_lot_fraction_percentage=80.0000%
      acquired_lot=InTransaction:
        id=4
        timestamp=2020-01-01 08:41:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=11000.0000
        crypto_in=1.00000000
        fiat_fee=100.0000
        fiat_in_no_fee=11000.0000
        fiat_in_with_fee=11100.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 2
      acquired_lot_fraction=2 of 2
      parent=3->None
    GainLoss:
      id=15->3
      crypto_amount=0.20000000
      fiat_cost_basis=2400.0000
      fiat_gain=40.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=2440.0000
      taxable_event_fraction_percentage=20.0000%
      taxable_event=OutTransaction:
        id=15
        timestamp=2020-02-11 19:58:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.SELL
        spot_price=12200.0000
        crypto_out_no_fee=1.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=12200.0000
      acquired_lot_fiat_amount_with_fee_fraction=2400.0000
      acquired_lot_fraction_percentage=10.0000%
      acquired_lot=InTransaction:
        id=3
        timestamp=2020-02-01 11:18:00.000000 +0000
        asset=B2
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=12000.0000
        crypto_in=2.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=24000.0000
        fiat_in_with_fee=24000.0000
        is_taxable=True
        fiat_taxable_amount=24000.0000
      taxable_event_fraction=2 of 2
      acquired_lot_fraction=1 of 2
      parent=15->4
    GainLoss:
      id=7->None
      crypto_amount=3.00000000
      fiat_cost_basis=0.0000
      fiat_gain=39000.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=39000.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=InTransaction:
        id=7
        timestamp=2020-03-01 09:45:00.000000 +0000
        asset=B2
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=13000.0000
        crypto_in=3.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=39000.0000
        fiat_in_with_fee=39000.0000
        is_taxable=True
        fiat_taxable_amount=39000.0000
      acquired_lot_fiat_amount_with_fee_fraction=0.0000
      acquired_lot_fraction_percentage=0.0000%
      acquired_lot=None
      taxable_event_fraction=1 of 1
      parent=15->3
    GainLoss:
      id=17->3
      crypto_amount=1.80000000
      fiat_cost_basis=21600.0000
      fiat_gain=3960.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=25560.0000
      taxable_event_fraction_percentage=36.0000%
      taxable_event=OutTransaction:
        id=17
        timestamp=2020-04-11 07:10:00.000000 +0000
        asset=B2
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.GIFT
        spot_price=14200.0000
        crypto_out_no_fee=5.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=71000.0000
      acquired_lot_fiat_amount_with_fee_fraction=21600.0000
      acquired_lot_fraction_percentage=90.0000%
      acquired_lot=InTransaction:
        id=3
        timestamp=2020-02-01 11:18:00.000000 +0000
        asset=B2
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=12000.0000
        crypto_in=2.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=24000.0000
        fiat_in_with_fee=24000.0000
        is_taxable=True
        fiat_taxable_amount=24000.0000
      taxable_event_fraction=1 of 3
      acquired_lot_fraction=2 of 2
      parent=7->None
    GainLoss:
      id=17->7
      crypto_amount=3.00000000
      fiat_cost_basis=39000.0000
      fiat_gain=3600.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=42600.0000
      taxable_event_fraction_percentage=60.0000%
      taxable_event=OutTransaction:
        id=17
        timestamp=2020-04-11 07:10:00.000000 +0000
        asset=B2
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.GIFT
        spot_price=14200.0000
        crypto_out_no_fee=5.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=71000.0000
      acquired_lot_fiat_amount_with_fee_fraction=39000.0000
      acquired_lot_fraction_percentage=100.0000%
      acquired_lot=InTransaction:
        id=7
        timestamp=2020-03-01 09:45:00.000000 +0000
        asset=B2
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=13000.0000
        crypto_in=3.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=39000.0000
        fiat_in_with_fee=39000.0000
        is_taxable=True
        fiat_taxable_amount=39000.0000
      taxable_event_fraction=2 of 3
      acquired_lot_fraction=1 of 1
      parent=17->3
    GainLoss:
      id=17->6
      crypto_amount=0.20000000
      fiat_cost_basis=2820.0000
      fiat_gain=20.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=2840.0000
      taxable_event_fraction_percentage=4.0000%
      taxable_event=OutTransaction:
        id=17
        timestamp=2020-04-11 07:10:00.000000 +0000
        asset=B2
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.GIFT
        spot_price=14200.0000
        crypto_out_no_fee=5.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=71000.0000
      acquired_lot_fiat_amount_with_fee_fraction=2820.0000
      acquired_lot_fraction_percentage=5.0000%
      acquired_lot=InTransaction:
        id=6
        timestamp=2020-04-01 09:45:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=14000.0000
        crypto_in=4.00000000
        fiat_fee=400.0000
        fiat_in_no_fee=56000.0000
        fiat_in_with_fee=56400.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=3 of 3
      acquired_lot_fraction=1 of 3
      parent=17->7
    GainLoss:
      id=13->6
      crypto_amount=3.79000000
      fiat_cost_basis=53439.0000
      fiat_gain=758.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=54197.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=OutTransaction:
        id=13
        timestamp=2020-04-12 17:50:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.DONATE
        spot_price=14300.0000
        crypto_out_no_fee=3.79000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=54197.0000
      acquired_lot_fiat_amount_with_fee_fraction=53439.0000
      acquired_lot_fraction_percentage=94.7500%
      acquired_lot=InTransaction:
        id=6
        timestamp=2020-04-01 09:45:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=14000.0000
        crypto_in=4.00000000
        fiat_fee=400.0000
        fiat_in_no_fee=56000.0000
        fiat_in_with_fee=56400.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=2 of 3
      parent=17->6
    GainLoss:
      id=14->6
      crypto_amount=0.01000000
      fiat_cost_basis=141.0000
      fiat_gain=59.9950
      is_long_term_capital_gains=True
      taxable_event_fiat_amount_with_fee_fraction=200.9950
      taxable_event_fraction_percentage=0.4975%
      taxable_event=OutTransaction:
        id=14
        timestamp=2021-06-11 05:31:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.SELL
        spot_price=20200.0000
        crypto_out_no_fee=2.00000000
        crypto_fee=0.01000000
        is_taxable=True
        fiat_taxable_amount=40400.0000
      acquired_lot_fiat_amount_with_fee_fraction=141.0000
      acquired_lot_fraction_percentage=0.2500%
      acquired_lot=InTransaction:
        id=6
        timestamp=2020-04-01 09:45:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=14000.0000
        crypto_in=4.00000000
        fiat_fee=400.0000
        fiat_in_no_fee=56000.0000
        fiat_in_with_fee=56400.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 2
      acquired_lot_fraction=3 of 3
      parent=13->6
    GainLoss:
      id=14->5
      crypto_amount=2.00000000
      fiat_cost_basis=30200.0000
      fiat_gain=9999.0050
      is_long_term_capital_gains=True
      taxable_event_fiat_amount_with_fee_fraction=40199.0050
      taxable_event_fraction_percentage=99.5025%
      taxable_event=OutTransaction:
        id=14
        timestamp=2021-06-11 05:31:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.SELL
        spot_price=20200.0000
        crypto_out_no_fee=2.00000000
        crypto_fee=0.01000000
        is_taxable=True
        fiat_taxable_amount=40400.0000
      acquired_lot_fiat_amount_with_fee_fraction=30200.0000
      acquired_lot_fraction_percentage=40.0000%
      acquired_lot=InTransaction:
        id=5
        timestamp=2020-05-01 14:03:00.000000 +0000
        asset=B2
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=15000.0000
        crypto_in=5.00000000
        fiat_fee=500.0000
        fiat_in_no_fee=75000.0000
        fiat_in_with_fee=75500.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=2 of 2
      acquired_lot_fraction=1 of 1
      parent=14->6""",
    "B3": """GainLossSet:
  configuration=./config/test_data.config
  asset=B3
  from_date=non-specified
  to_date=non-specified
  entries=
    GainLoss:
      id=26->4
      crypto_amount=0.01000000
      fiat_cost_basis=111.0000
      fiat_gain=3.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=114.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=IntraTransaction:
        id=26
        timestamp=2020-01-21 18:33:14.342000 +0000
        asset=B3
        from_exchange=Coinbase
        from_holder=Bob
        to_exchange=BlockFi
        to_holder=Bob
        transaction_type=TransactionType.MOVE
        spot_price=11400.0000
        crypto_sent=0.10000000
        crypto_received=0.09000000
        crypto_fee=0.01000000
        fiat_fee=114.0000
        is_taxable=True
        fiat_taxable_amount=114.0000
      acquired_lot_fiat_amount_with_fee_fraction=111.0000
      acquired_lot_fraction_percentage=1.0000%
      acquired_lot=InTransaction:
        id=4
        timestamp=2020-01-01 08:41:00.000000 +0000
        asset=B3
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=11000.0000
        crypto_in=1.00000000
        fiat_fee=100.0000
        fiat_in_no_fee=11000.0000
        fiat_in_with_fee=11100.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=1 of 3
      parent=None
    GainLoss:
      id=3->None
      crypto_amount=2.00000000
      fiat_cost_basis=0.0000
      fiat_gain=24000.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=24000.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=InTransaction:
        id=3
        timestamp=2020-02-01 11:18:00.000000 +0000
        asset=B3
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=12000.0000
        crypto_in=2.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=24000.0000
        fiat_in_with_fee=24000.0000
        is_taxable=True
        fiat_taxable_amount=24000.0000
      acquired_lot_fiat_amount_with_fee_fraction=0.0000
      acquired_lot_fraction_percentage=0.0000%
      acquired_lot=None
      taxable_event_fraction=1 of 1
      parent=26->4
    GainLoss:
      id=7->None
      crypto_amount=3.00000000
      fiat_cost_basis=0.0000
      fiat_gain=39000.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=39000.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=InTransaction:
        id=7
        timestamp=2020-03-01 09:45:00.000000 +0000
        asset=B3
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=13000.0000
        crypto_in=3.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=39000.0000
        fiat_in_with_fee=39000.0000
        is_taxable=True
        fiat_taxable_amount=39000.0000
      acquired_lot_fiat_amount_with_fee_fraction=0.0000
      acquired_lot_fraction_percentage=0.0000%
      acquired_lot=None
      taxable_event_fraction=1 of 1
      parent=3->None
    GainLoss:
      id=25->4
      crypto_amount=0.02000000
      fiat_cost_basis=222.0000
      fiat_gain=66.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=288.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=IntraTransaction:
        id=25
        timestamp=2020-05-21 12:58:10.000000 +0000
        asset=B3
        from_exchange=Coinbase
        from_holder=Bob
        to_exchange=Kraken
        to_holder=Alice
        transaction_type=TransactionType.MOVE
        spot_price=14400.0000
        crypto_sent=0.20000000
        crypto_received=0.18000000
        crypto_fee=0.02000000
        fiat_fee=288.0000
        is_taxable=True
        fiat_taxable_amount=288.0000
      acquired_lot_fiat_amount_with_fee_fraction=222.0000
      acquired_lot_fraction_percentage=2.0000%
      acquired_lot=InTransaction:
        id=4
        timestamp=2020-01-01 08:41:00.000000 +0000
        asset=B3
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=11000.0000
        crypto_in=1.00000000
        fiat_fee=100.0000
        fiat_in_no_fee=11000.0000
        fiat_in_with_fee=11100.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=2 of 3
      parent=7->None
    GainLoss:
      id=23->4
      crypto_amount=0.04000000
      fiat_cost_basis=444.0000
      fiat_gain=412.0000
      is_long_term_capital_gains=True
      taxable_event_fiat_amount_with_fee_fraction=856.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=IntraTransaction:
        id=23
        timestamp=2021-07-21 10:02:02.000000 +0000
        asset=B3
        from_exchange=Coinbase
        from_holder=Bob
        to_exchange=Kraken
        to_holder=Alice
        transaction_type=TransactionType.MOVE
        spot_price=21400.0000
        crypto_sent=0.50000000
        crypto_received=0.46000000
        crypto_fee=0.04000000
        fiat_fee=856.0000
        is_taxable=True
        fiat_taxable_amount=856.0000
      acquired_lot_fiat_amount_with_fee_fraction=444.0000
      acquired_lot_fraction_percentage=4.0000%
      acquired_lot=InTransaction:
        id=4
        timestamp=2020-01-01 08:41:00.000000 +0000
        asset=B3
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=11000.0000
        crypto_in=1.00000000
        fiat_fee=100.0000
        fiat_in_no_fee=11000.0000
        fiat_in_with_fee=11100.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=3 of 3
      parent=25->4""",
    "B4": """GainLossSet:
  configuration=./config/test_data.config
  asset=B4
  from_date=non-specified
  to_date=non-specified
  entries=
    GainLoss:
      id=16->4
      crypto_amount=0.20000000
      fiat_cost_basis=2220.0000
      fiat_gain=20.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=2240.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=OutTransaction:
        id=16
        timestamp=2020-01-11 11:15:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.SELL
        spot_price=11200.0000
        crypto_out_no_fee=0.20000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=2240.0000
      acquired_lot_fiat_amount_with_fee_fraction=2220.0000
      acquired_lot_fraction_percentage=20.0000%
      acquired_lot=InTransaction:
        id=4
        timestamp=2020-01-01 08:41:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=11000.0000
        crypto_in=1.00000000
        fiat_fee=100.0000
        fiat_in_no_fee=11000.0000
        fiat_in_with_fee=11100.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=1 of 3
      parent=None
    GainLoss:
      id=26->4
      crypto_amount=0.01000000
      fiat_cost_basis=111.0000
      fiat_gain=3.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=114.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=IntraTransaction:
        id=26
        timestamp=2020-01-21 18:33:14.342000 +0000
        asset=B4
        from_exchange=Coinbase
        from_holder=Bob
        to_exchange=BlockFi
        to_holder=Bob
        transaction_type=TransactionType.MOVE
        spot_price=11400.0000
        crypto_sent=0.10000000
        crypto_received=0.09000000
        crypto_fee=0.01000000
        fiat_fee=114.0000
        is_taxable=True
        fiat_taxable_amount=114.0000
      acquired_lot_fiat_amount_with_fee_fraction=111.0000
      acquired_lot_fraction_percentage=1.0000%
      acquired_lot=InTransaction:
        id=4
        timestamp=2020-01-01 08:41:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=11000.0000
        crypto_in=1.00000000
        fiat_fee=100.0000
        fiat_in_no_fee=11000.0000
        fiat_in_with_fee=11100.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=2 of 3
      parent=16->4
    GainLoss:
      id=3->None
      crypto_amount=2.00000000
      fiat_cost_basis=0.0000
      fiat_gain=24000.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=24000.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=InTransaction:
        id=3
        timestamp=2020-02-01 11:18:00.000000 +0000
        asset=B4
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=12000.0000
        crypto_in=2.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=24000.0000
        fiat_in_with_fee=24000.0000
        is_taxable=True
        fiat_taxable_amount=24000.0000
      acquired_lot_fiat_amount_with_fee_fraction=0.0000
      acquired_lot_fraction_percentage=0.0000%
      acquired_lot=None
      taxable_event_fraction=1 of 1
      parent=26->4
    GainLoss:
      id=15->4
      crypto_amount=0.79000000
      fiat_cost_basis=8769.0000
      fiat_gain=869.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=9638.0000
      taxable_event_fraction_percentage=79.0000%
      taxable_event=OutTransaction:
        id=15
        timestamp=2020-02-11 19:58:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.SELL
        spot_price=12200.0000
        crypto_out_no_fee=1.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=12200.0000
      acquired_lot_fiat_amount_with_fee_fraction=8769.0000
      acquired_lot_fraction_percentage=79.0000%
      acquired_lot=InTransaction:
        id=4
        timestamp=2020-01-01 08:41:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=11000.0000
        crypto_in=1.00000000
        fiat_fee=100.0000
        fiat_in_no_fee=11000.0000
        fiat_in_with_fee=11100.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 2
      acquired_lot_fraction=3 of 3
      parent=3->None
    GainLoss:
      id=15->3
      crypto_amount=0.21000000
      fiat_cost_basis=2520.0000
      fiat_gain=42.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=2562.0000
      taxable_event_fraction_percentage=21.0000%
      taxable_event=OutTransaction:
        id=15
        timestamp=2020-02-11 19:58:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.SELL
        spot_price=12200.0000
        crypto_out_no_fee=1.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=12200.0000
      acquired_lot_fiat_amount_with_fee_fraction=2520.0000
      acquired_lot_fraction_percentage=10.5000%
      acquired_lot=InTransaction:
        id=3
        timestamp=2020-02-01 11:18:00.000000 +0000
        asset=B4
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=12000.0000
        crypto_in=2.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=24000.0000
        fiat_in_with_fee=24000.0000
        is_taxable=True
        fiat_taxable_amount=24000.0000
      taxable_event_fraction=2 of 2
      acquired_lot_fraction=1 of 2
      parent=15->4
    GainLoss:
      id=7->None
      crypto_amount=3.00000000
      fiat_cost_basis=0.0000
      fiat_gain=39000.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=39000.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=InTransaction:
        id=7
        timestamp=2020-03-01 09:45:00.000000 +0000
        asset=B4
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=13000.0000
        crypto_in=3.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=39000.0000
        fiat_in_with_fee=39000.0000
        is_taxable=True
        fiat_taxable_amount=39000.0000
      acquired_lot_fiat_amount_with_fee_fraction=0.0000
      acquired_lot_fraction_percentage=0.0000%
      acquired_lot=None
      taxable_event_fraction=1 of 1
      parent=15->3
    GainLoss:
      id=17->3
      crypto_amount=1.79000000
      fiat_cost_basis=21480.0000
      fiat_gain=3938.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=25418.0000
      taxable_event_fraction_percentage=35.8000%
      taxable_event=OutTransaction:
        id=17
        timestamp=2020-04-11 07:10:00.000000 +0000
        asset=B4
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.GIFT
        spot_price=14200.0000
        crypto_out_no_fee=5.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=71000.0000
      acquired_lot_fiat_amount_with_fee_fraction=21480.0000
      acquired_lot_fraction_percentage=89.5000%
      acquired_lot=InTransaction:
        id=3
        timestamp=2020-02-01 11:18:00.000000 +0000
        asset=B4
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=12000.0000
        crypto_in=2.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=24000.0000
        fiat_in_with_fee=24000.0000
        is_taxable=True
        fiat_taxable_amount=24000.0000
      taxable_event_fraction=1 of 3
      acquired_lot_fraction=2 of 2
      parent=7->None
    GainLoss:
      id=17->7
      crypto_amount=3.00000000
      fiat_cost_basis=39000.0000
      fiat_gain=3600.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=42600.0000
      taxable_event_fraction_percentage=60.0000%
      taxable_event=OutTransaction:
        id=17
        timestamp=2020-04-11 07:10:00.000000 +0000
        asset=B4
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.GIFT
        spot_price=14200.0000
        crypto_out_no_fee=5.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=71000.0000
      acquired_lot_fiat_amount_with_fee_fraction=39000.0000
      acquired_lot_fraction_percentage=100.0000%
      acquired_lot=InTransaction:
        id=7
        timestamp=2020-03-01 09:45:00.000000 +0000
        asset=B4
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.INTEREST
        spot_price=13000.0000
        crypto_in=3.00000000
        fiat_fee=0.0000
        fiat_in_no_fee=39000.0000
        fiat_in_with_fee=39000.0000
        is_taxable=True
        fiat_taxable_amount=39000.0000
      taxable_event_fraction=2 of 3
      acquired_lot_fraction=1 of 1
      parent=17->3
    GainLoss:
      id=17->6
      crypto_amount=0.21000000
      fiat_cost_basis=2961.0000
      fiat_gain=21.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=2982.0000
      taxable_event_fraction_percentage=4.2000%
      taxable_event=OutTransaction:
        id=17
        timestamp=2020-04-11 07:10:00.000000 +0000
        asset=B4
        exchange=BlockFi
        holder=Bob
        transaction_type=TransactionType.GIFT
        spot_price=14200.0000
        crypto_out_no_fee=5.00000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=71000.0000
      acquired_lot_fiat_amount_with_fee_fraction=2961.0000
      acquired_lot_fraction_percentage=5.2500%
      acquired_lot=InTransaction:
        id=6
        timestamp=2020-04-01 09:45:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=14000.0000
        crypto_in=4.00000000
        fiat_fee=400.0000
        fiat_in_no_fee=56000.0000
        fiat_in_with_fee=56400.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=3 of 3
      acquired_lot_fraction=1 of 2
      parent=17->7
    GainLoss:
      id=13->6
      crypto_amount=3.79000000
      fiat_cost_basis=53439.0000
      fiat_gain=758.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=54197.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=OutTransaction:
        id=13
        timestamp=2020-04-12 17:50:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.DONATE
        spot_price=14300.0000
        crypto_out_no_fee=3.79000000
        crypto_fee=0.00000000
        is_taxable=True
        fiat_taxable_amount=54197.0000
      acquired_lot_fiat_amount_with_fee_fraction=53439.0000
      acquired_lot_fraction_percentage=94.7500%
      acquired_lot=InTransaction:
        id=6
        timestamp=2020-04-01 09:45:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=14000.0000
        crypto_in=4.00000000
        fiat_fee=400.0000
        fiat_in_no_fee=56000.0000
        fiat_in_with_fee=56400.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=2 of 2
      parent=17->6
    GainLoss:
      id=25->5
      crypto_amount=0.02000000
      fiat_cost_basis=302.0000
      fiat_gain=-14.0000
      is_long_term_capital_gains=False
      taxable_event_fiat_amount_with_fee_fraction=288.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=IntraTransaction:
        id=25
        timestamp=2020-05-21 12:58:10.000000 +0000
        asset=B4
        from_exchange=Coinbase
        from_holder=Bob
        to_exchange=Kraken
        to_holder=Alice
        transaction_type=TransactionType.MOVE
        spot_price=14400.0000
        crypto_sent=0.20000000
        crypto_received=0.18000000
        crypto_fee=0.02000000
        fiat_fee=288.0000
        is_taxable=True
        fiat_taxable_amount=288.0000
      acquired_lot_fiat_amount_with_fee_fraction=302.0000
      acquired_lot_fraction_percentage=0.4000%
      acquired_lot=InTransaction:
        id=5
        timestamp=2020-05-01 14:03:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=15000.0000
        crypto_in=5.00000000
        fiat_fee=500.0000
        fiat_in_no_fee=75000.0000
        fiat_in_with_fee=75500.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=1 of 3
      parent=13->6
    GainLoss:
      id=14->5
      crypto_amount=2.01000000
      fiat_cost_basis=30351.0000
      fiat_gain=10049.0000
      is_long_term_capital_gains=True
      taxable_event_fiat_amount_with_fee_fraction=40400.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=OutTransaction:
        id=14
        timestamp=2021-06-11 05:31:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.SELL
        spot_price=20200.0000
        crypto_out_no_fee=2.00000000
        crypto_fee=0.01000000
        is_taxable=True
        fiat_taxable_amount=40400.0000
      acquired_lot_fiat_amount_with_fee_fraction=30351.0000
      acquired_lot_fraction_percentage=40.2000%
      acquired_lot=InTransaction:
        id=5
        timestamp=2020-05-01 14:03:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=15000.0000
        crypto_in=5.00000000
        fiat_fee=500.0000
        fiat_in_no_fee=75000.0000
        fiat_in_with_fee=75500.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=2 of 3
      parent=25->5
    GainLoss:
      id=23->5
      crypto_amount=0.04000000
      fiat_cost_basis=604.0000
      fiat_gain=252.0000
      is_long_term_capital_gains=True
      taxable_event_fiat_amount_with_fee_fraction=856.0000
      taxable_event_fraction_percentage=100.0000%
      taxable_event=IntraTransaction:
        id=23
        timestamp=2021-07-21 10:02:02.000000 +0000
        asset=B4
        from_exchange=Coinbase
        from_holder=Bob
        to_exchange=Kraken
        to_holder=Alice
        transaction_type=TransactionType.MOVE
        spot_price=21400.0000
        crypto_sent=0.50000000
        crypto_received=0.46000000
        crypto_fee=0.04000000
        fiat_fee=856.0000
        is_taxable=True
        fiat_taxable_amount=856.0000
      acquired_lot_fiat_amount_with_fee_fraction=604.0000
      acquired_lot_fraction_percentage=0.8000%
      acquired_lot=InTransaction:
        id=5
        timestamp=2020-05-01 14:03:00.000000 +0000
        asset=B4
        exchange=Coinbase
        holder=Bob
        transaction_type=TransactionType.BUY
        spot_price=15000.0000
        crypto_in=5.00000000
        fiat_fee=500.0000
        fiat_in_no_fee=75000.0000
        fiat_in_with_fee=75500.0000
        is_taxable=False
        fiat_taxable_amount=0.0000
      taxable_event_fraction=1 of 1
      acquired_lot_fraction=3 of 3
      parent=14->5""",
}



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Callable, List, Optional

from rp2.abstract_entry import AbstractEntry
from rp2.configuration import Configuration
from rp2.entry_types import TransactionType
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError


class AbstractTransaction(AbstractEntry):
    def __init__(
        self,
        configuration: Configuration,
        timestamp: str,
        asset: str,
        transaction_type: str,
        spot_price: RP2Decimal,
        unique_id: Optional[int] = None,
        notes: Optional[str] = None,
    ) -> None:
        super().__init__(configuration, asset)

        self.__timestamp: datetime = configuration.type_check_timestamp_from_string("timestamp", timestamp)
        self.__transaction_type: TransactionType = TransactionType.type_check_from_string("transaction_type", transaction_type)
        self.__spot_price: RP2Decimal = configuration.type_check_positive_decimal("spot_price", spot_price)
        self.__unique_id: int = configuration.type_check_unique_id("unique_id", unique_id) if unique_id is not None else id(self)
        self.__notes = configuration.type_check_string("notes", notes) if notes else ""

    @classmethod
    def type_check(cls, name: str, instance: "AbstractEntry") -> "AbstractEntry":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __eq__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, AbstractTransaction):
            raise RP2TypeError(f"Operand has non-AbstractTransaction value {repr(other)}")
        # By definition, unique_id can uniquely identify a transaction: this works even if it's the ODS line from the spreadsheet,
        # since there are no cross-asset transactions (so a spreadsheet line points to a unique transaction for that asset).
        result: bool = self.unique_id == other.unique_id
        return result

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        # By definition, unique_id can uniquely identify a transaction: this works even if it's the ODS line from the spreadsheet,
        # since there are no cross-asset transactions (so a spreadsheet line points to a unique transaction for that asset).
        return hash(self.unique_id)

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:
        class_specific_data: List[str] = []
        stringify: Callable[[object], str] = repr
        if not repr_format:
            stringify = str

        class_specific_data.append(f"timestamp={stringify(self.timestamp.strftime('%Y-%m-%d %H:%M:%S.%f %z'))}")
        class_specific_data.append(f"asset={stringify(self.asset)}")

        if extra_data:
            class_specific_data.extend(extra_data)

        return super().to_string(indent=indent, repr_format=repr_format, extra_data=class_specific_data)

    @property
    def unique_id(self) -> str:
        return str(self.__unique_id)

    @property
    def timestamp(self) -> datetime:
        return self.__timestamp

    @property
    def transaction_type(self) -> TransactionType:
        return self.__transaction_type

    @property
    def spot_price(self) -> RP2Decimal:
        return self.__spot_price

    @property
    def notes(self) -> str:
        return self.__notes

    # Crypto amount that is taxed
    @property
    def crypto_taxable_amount(self) -> RP2Decimal:
        raise NotImplementedError("Abstract property")

    # Fiat amount that is taxed
    @property
    def fiat_taxable_amount(self) -> RP2Decimal:
        raise NotImplementedError("Abstract property")

    # Crypto amount that is not taxed
    @property
    def crypto_deduction(self) -> RP2Decimal:
        raise NotImplementedError("Abstract property")

    # Fiat amount that is not taxed
    @property
    def fiat_deduction(self) -> RP2Decimal:
        raise NotImplementedError("Abstract property")

    def is_taxable(self) -> bool:
        raise NotImplementedError("Abstract method")



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from rp2.configuration import Configuration
from rp2.gain_loss import GainLoss
from rp2.in_transaction import InTransaction
from rp2.intra_transaction import IntraTransaction
from rp2.out_transaction import OutTransaction
from rp2.plugin.accounting_method.fifo import AccountingMethod
from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError


class TestGainLoss(unittest.TestCase):
    # pylint: disable=line-too-long
    _configuration: Configuration
    _accounting_method: AccountingMethod

    @classmethod
    def setUpClass(cls) -> None:
        cls._configuration = Configuration("./config/test_data.config", US())
        cls._accounting_method = AccountingMethod()

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

        self._in_buy = InTransaction(
            self._configuration,
            "2020-01-02T08:42:43.882Z",
            "B1",
            "Coinbase Pro",
            "Bob",
            "BuY",
            RP2Decimal("10000"),
            RP2Decimal("2.0002"),
            RP2Decimal("20"),
            RP2Decimal("20002"),
            RP2Decimal("20022"),
            unique_id=10,
        )
        self._in_buy2 = InTransaction(
            self._configuration,
            "2020-01-12T17:33:18Z",
            "B1",
            "Coinbase Pro",
            "Bob",
            "BuY",
            RP2Decimal("10500"),
            RP2Decimal("0.8"),
            RP2Decimal("10"),
            unique_id=11,
        )
        self._in_buy3 = InTransaction(
            self._configuration,
            "2020-04-27T03:28:47Z",
            "B1",
            "Coinbase Pro",
            "Bob",
            "BuY",
            RP2Decimal("1300"),
            RP2Decimal("1.5"),
            RP2Decimal("20"),
            unique_id=12,
        )
        self._in_interest = InTransaction(
            self._configuration,
            "2020-02-21T13:14:08 -00:04",
            "B1",
            "BlockFi",
            "Bob",
            "interest",
            RP2Decimal("11000"),
            RP2Decimal("0.1"),
            RP2Decimal("0"),
            unique_id=14,
        )
        self._out: OutTransaction = OutTransaction(
            self._configuration,
            "3/3/2020 3:59:59 -04:00",
            "B1",
            "Coinbase Pro",
            "Bob",
            "SELL",
            RP2Decimal("12000"),
            RP2Decimal("0.2"),
            RP2Decimal("0"),
            unique_id=20,
        )
        self._intra: IntraTransaction = IntraTransaction(
            self._configuration,
            "2021-03-10T11:18:58 -00:04",
            "B1",
            "Coinbase Pro",
            "Bob",
            "BlockFi",
            "Alice",
            RP2Decimal("12500.0"),
            RP2Decimal("0.4"),
            RP2Decimal("0.39"),
            unique_id=30,
        )

    def test_good_interest_gain_loss(self) -> None:
        flow: GainLoss = GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.1"), self._in_interest, None)
        self.assertEqual(flow.crypto_amount, RP2Decimal("0.1"))
        self.assertEqual(flow.taxable_event, self._in_interest)
        self.assertEqual(flow.acquired_lot, None)
        self.assertEqual(flow.timestamp, flow.taxable_event.timestamp)
        self.assertEqual(flow.crypto_balance_change, RP2Decimal("0.1"))
        self.assertEqual(flow.taxable_event_fiat_amount_with_fee_fraction, RP2Decimal("1100"))
        self.assertEqual(
            str(flow),
            """GainLoss:
  id=14->None
  crypto_amount=0.10000000
  fiat_cost_basis=0.0000
  fiat_gain=1100.0000
  is_long_term_capital_gains=False
  taxable_event_fiat_amount_with_fee_fraction=1100.0000
  taxable_event_fraction_percentage=100.0000%
  taxable_event=InTransaction:
    id=14
    timestamp=2020-02-21 13:14:08.000000 -0004
    asset=B1
    exchange=BlockFi
    holder=Bob
    transaction_type=TransactionType.INTEREST
    spot_price=11000.0000
    crypto_in=0.10000000
    fiat_fee=0.0000
    fiat_in_no_fee=1100.0000
    fiat_in_with_fee=1100.0000
    is_taxable=True
    fiat_taxable_amount=1100.0000
  acquired_lot_fiat_amount_with_fee_fraction=0.0000
  acquired_lot_fraction_percentage=0.0000%
  acquired_lot=None""",
        )
        self.assertEqual(
            repr(flow),
            "GainLoss(id='14->None', crypto_amount=0.10000000, fiat_cost_basis=0.0000, fiat_gain=1100.0000, is_long_term_capital_gains=False, taxable_event_fiat_amount_with_fee_fraction=1100.0000, taxable_event_fraction_percentage=100.0000%, taxable_event=InTransaction(id='14', timestamp='2020-02-21 13:14:08.000000 -0004', asset='B1', exchange='BlockFi', holder='Bob', transaction_type=<TransactionType.INTEREST: 'interest'>, spot_price=11000.0000, crypto_in=0.10000000, fiat_fee=0.0000, fiat_in_no_fee=1100.0000, fiat_in_with_fee=1100.0000, is_taxable=True, fiat_taxable_amount=1100.0000), acquired_lot_fiat_amount_with_fee_fraction=0.0000, acquired_lot_fraction_percentage=0.0000%, acquired_lot=None)",
        )

    def test_good_non_interest_gain_loss(self) -> None:
        flow: GainLoss = GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.001"), self._intra, self._in_buy)
        self.assertEqual(flow.crypto_amount, RP2Decimal("0.001"))
        self.assertEqual(flow.taxable_event, self._intra)
        self.assertEqual(flow.acquired_lot, self._in_buy)
        self.assertEqual(flow.timestamp, flow.taxable_event.timestamp)
        self.assertEqual(flow.crypto_balance_change, RP2Decimal("0.001"))
        self.assertEqual(flow.taxable_event_fiat_amount_with_fee_fraction, RP2Decimal("12.5"))
        self.assertEqual(
            str(flow),
            """GainLoss:
  id=30->10
  crypto_amount=0.00100000
  fiat_cost_basis=10.0100
  fiat_gain=2.4900
  is_long_term_capital_gains=True
  taxable_event_fiat_amount_with_fee_fraction=12.5000
  taxable_event_fraction_percentage=10.0000%
  taxable_event=IntraTransaction:
    id=30
    timestamp=2021-03-10 11:18:58.000000 -0004
    asset=B1
    from_exchange=Coinbase Pro
    from_holder=Bob
    to_exchange=BlockFi
    to_holder=Alice
    transaction_type=TransactionType.MOVE
    spot_price=12500.0000
    crypto_sent=0.40000000
    crypto_received=0.39000000
    crypto_fee=0.01000000
    fiat_fee=125.0000
    is_taxable=True
    fiat_taxable_amount=125.0000
  acquired_lot_fiat_amount_with_fee_fraction=10.0100
  acquired_lot_fraction_percentage=0.0500%
  acquired_lot=InTransaction:
    id=10
    timestamp=2020-01-02 08:42:43.882000 +0000
    asset=B1
    exchange=Coinbase Pro
    holder=Bob
    transaction_type=TransactionType.BUY
    spot_price=10000.0000
    crypto_in=2.00020000
    fiat_fee=20.0000
    fiat_in_no_fee=20002.0000
    fiat_in_with_fee=20022.0000
    is_taxable=False
    fiat_taxable_amount=0.0000""",
        )
        self.assertEqual(
            repr(flow),
            "GainLoss(id='30->10', crypto_amount=0.00100000, fiat_cost_basis=10.0100, fiat_gain=2.4900, is_long_term_capital_gains=True, taxable_event_fiat_amount_with_fee_fraction=12.5000, taxable_event_fraction_percentage=10.0000%, taxable_event=IntraTransaction(id='30', timestamp='2021-03-10 11:18:58.000000 -0004', asset='B1', from_exchange='Coinbase Pro', from_holder='Bob', to_exchange='BlockFi', to_holder='Alice', transaction_type=<TransactionType.MOVE: 'move'>, spot_price=12500.0000, crypto_sent=0.40000000, crypto_received=0.39000000, crypto_fee=0.01000000, fiat_fee=125.0000, is_taxable=True, fiat_taxable_amount=125.0000), acquired_lot_fiat_amount_with_fee_fraction=10.0100, acquired_lot_fraction_percentage=0.0500%, acquired_lot=InTransaction(id='10', timestamp='2020-01-02 08:42:43.882000 +0000', asset='B1', exchange='Coinbase Pro', holder='Bob', transaction_type=<TransactionType.BUY: 'buy'>, spot_price=10000.0000, crypto_in=2.00020000, fiat_fee=20.0000, fiat_in_no_fee=20002.0000, fiat_in_with_fee=20022.0000, is_taxable=False, fiat_taxable_amount=0.0000))",
        )

    def test_gain_loss_equality_and_hashing(self) -> None:
        gain_loss: GainLoss = GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.001"), self._intra, self._in_buy)
        gain_loss2: GainLoss = GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.001"), self._intra, self._in_buy)
        gain_loss3: GainLoss = GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.001"), self._intra, self._in_buy2)
        gain_loss4: GainLoss = GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.001"), self._out, self._in_buy)
        gain_loss5: GainLoss = GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.001"), self._out, self._in_buy2)
        gain_loss6: GainLoss = GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.1"), self._in_interest, None)
        self.assertEqual(gain_loss, gain_loss)
        self.assertEqual(gain_loss, gain_loss2)
        self.assertNotEqual(gain_loss, gain_loss3)
        self.assertNotEqual(gain_loss, gain_loss4)
        self.assertNotEqual(gain_loss, gain_loss5)
        self.assertNotEqual(gain_loss, gain_loss6)
        self.assertEqual(hash(gain_loss), hash(gain_loss))
        self.assertEqual(hash(gain_loss), hash(gain_loss2))
        # These hashes would only be equal in case of hash collision (possible but very unlikey).
        self.assertNotEqual(hash(gain_loss), hash(gain_loss3))
        self.assertNotEqual(hash(gain_loss), hash(gain_loss4))
        self.assertNotEqual(hash(gain_loss), hash(gain_loss5))
        self.assertNotEqual(hash(gain_loss), hash(gain_loss6))

    def test_bad_gain_loss(self) -> None:
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            GainLoss(None, self._accounting_method, RP2Decimal("0.5"), self._in_interest, None)  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            GainLoss("config", self._accounting_method, RP2Decimal("0.5"), self._in_interest, None)  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'accounting_method' is not of type AbstractAccountingMethod: .*"):
            # Bad configuration
            GainLoss(self._configuration, None, RP2Decimal("0.5"), self._in_interest, None)  # type: ignore

        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_amount' has non-positive value .*"):
            # Bad amount
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("-1"), self._out, None)

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'crypto_amount' has non-RP2Decimal value"):
            # Bad amount
            GainLoss(self._configuration, self._accounting_method, "0.5", self._in_interest, None)  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'taxable_event' is not of type AbstractTransaction: .*"):
            # Bad taxable event
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.5"), None, self._in_buy)  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'taxable_event' is not of type AbstractTransaction: .*"):
            # Bad taxable event
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.5"), "foobar", self._in_buy)  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'acquired_lot' is not of type InTransaction: "):
            # Bad acquired lot
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.1"), self._out, 33)  # type: ignore

        with self.assertRaisesRegex(
            RP2TypeError,
            "acquired_lot must be None for earn-typed taxable_events, instead it's foobar",
        ):
            # Bad acquired lot
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.1"), self._in_interest, "foobar")  # type: ignore

        with self.assertRaisesRegex(RP2ValueError, "Parameter 'taxable_event' of class InTransaction is not taxable: .*"):
            # Taxable event not taxable
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.2"), self._in_buy2, self._in_buy)

        with self.assertRaisesRegex(
            RP2ValueError,
            "crypto_amount must be == taxable_event.crypto_balance_change for earn-typed taxable events, but they differ .* != .*",
        ):
            # Earn-typed taxable event: acquired_lot not None
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("1.1"), self._in_interest, None)

        with self.assertRaisesRegex(
            RP2TypeError,
            "acquired_lot must be None for earn-typed taxable_events, instead it's .*",
        ):
            # Earn-typed taxable event: acquired_lot not None
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.1"), self._in_interest, self._in_buy2)

        with self.assertRaisesRegex(RP2TypeError, "acquired_lot must not be None for non-earn-typed taxable_events"):
            # Non-earn-typed taxable event: acquired lot None
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.2"), self._out, None)

        with self.assertRaisesRegex(
            RP2ValueError,
            "crypto_amount .* is greater than taxable event amount .* or acquired-lot amount .*: ",
        ):
            # Non-earn-typed taxable event: acquired_lot not None
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("2"), self._out, self._in_buy2)

        with self.assertRaisesRegex(RP2ValueError, "Timestamp .* of taxable_event is earlier than timestamp .* of acquired_lot: .*"):
            # Non-earn-typed taxable event: acquired_lot not None
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.1"), self._out, self._in_buy3)

        with self.assertRaisesRegex(RP2ValueError, "taxable_event.asset .* != acquired_lot.asset .*"):
            # Mix different assets (B1 and B2) in the same GainLoss
            in_transaction: InTransaction = InTransaction(
                self._configuration,
                "2019-04-27T03:28:47Z",
                "B2",
                "Coinbase Pro",
                "Bob",
                "BuY",
                RP2Decimal("1300"),
                RP2Decimal("1.5"),
                RP2Decimal("20"),
                unique_id=11,
            )
            GainLoss(self._configuration, self._accounting_method, RP2Decimal("0.1"), self._out, in_transaction)


if __name__ == "__main__":
    unittest.main()
